# -*- coding: utf-8 -*-

import itertools
import logging
import operator
from collections import defaultdict, MutableMapping
from functools import reduce

from django.db import models
from django.db.models import Manager
from django.contrib.contenttypes.fields import GenericRelation, GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ValidationError, ObjectDoesNotExist
from django.contrib.postgres.fields import HStoreField, JSONField, RangeField
from django.utils.translation import ungettext

from neomodel import *
from chemtrails.conf import settings
from chemtrails.utils import get_model_string, flatten, timeit

logger = logging.getLogger(__name__)

field_property_map = {
    models.ForeignKey: RelationshipTo,
    models.ManyToOneRel: RelationshipTo,
    models.OneToOneField: RelationshipTo,
    models.OneToOneRel: RelationshipTo,
    models.ManyToManyField: RelationshipTo,
    models.ManyToManyRel: RelationshipTo,
    GenericForeignKey: RelationshipTo,
    GenericRelation: RelationshipTo,

    models.AutoField: IntegerProperty,
    models.BigIntegerField: IntegerProperty,
    models.BooleanField: BooleanProperty,
    models.CharField: StringProperty,
    models.CommaSeparatedIntegerField: ArrayProperty,
    models.DateField: DateProperty,
    models.DateTimeField: DateTimeProperty,
    models.DecimalField: FloatProperty,
    models.DurationField: StringProperty,
    models.EmailField: StringProperty,
    models.FilePathField: StringProperty,
    models.FileField: StringProperty,
    models.FloatField: FloatProperty,
    models.GenericIPAddressField: StringProperty,
    models.IntegerField: IntegerProperty,
    models.IPAddressField: StringProperty,
    models.NullBooleanField: BooleanProperty,
    models.PositiveIntegerField: IntegerProperty,
    models.PositiveSmallIntegerField: IntegerProperty,
    models.SlugField: StringProperty,
    models.SmallIntegerField: IntegerProperty,
    models.TextField: StringProperty,
    models.TimeField: IntegerProperty,
    models.URLField: StringProperty,
    models.UUIDField: StringProperty,

    # PostgreSQL special fields
    HStoreField: JSONProperty,
    JSONField: JSONProperty
}


class DefinitionDict(MutableMapping):
    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))
        self._hidden = {}

    def __getitem__(self, key):
        try:
            return self._hidden[key]
        except KeyError:
            return self.store[key]

    def __delitem__(self, key):
        del self.store[key]

    def __setitem__(self, key, value):
        self.store[key] = value

    def items(self):
        _items = self.store.items()
        _r = []
        for a,b in _items:
            if a in self._hidden:
                _r.append([a, self._hidden[a]])
            else:
                _r.append([a,b])

        return _r

    def _safeguard(self, key, value):
        self._hidden[key] = value

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)


class Meta(type):
    """
    Meta class template.
    """
    model = None
    app_label = None

    def __new__(mcs, name, bases, attrs):
        cls = super(Meta, mcs).__new__(mcs, str(name), bases, attrs)
        return cls


class NodeBase(NodeMeta):
    """
    Base Meta class for ``StructuredNode`` which adds a model class.
    """
    creation_counter = 0

    def __new__(mcs, name, bases, attrs):
        cls = super(NodeBase, mcs).__new__(mcs, str(name), bases, attrs)
        cls.creation_counter = NodeBase.creation_counter
        NodeBase.creation_counter += 1

        if getattr(cls, 'Meta', None):
            meta = Meta('Meta', (Meta,), dict(cls.Meta.__dict__))

            # A little hack which helps us dynamically create ModelNode classes
            # where variables holding the model class is out of scope.
            if hasattr(cls, '__metaclass_model__') and not meta.model:
                meta.model = getattr(cls, '__metaclass_model__', None)
                delattr(cls, '__metaclass_model__')

            if not getattr(meta, 'model', None):
                raise ValueError('%s.Meta.model attribute cannot be None.' % name)

            if getattr(meta, 'app_label', None) is None:
                meta.app_label = meta.model._meta.app_label

            cls.add_to_class('Meta', meta)

        elif not getattr(cls, '__abstract_node__', None):
            raise ImproperlyConfigured('%s must implement a Meta class.' % name)

        return cls

    def add_to_class(cls, name, value):
        setattr(cls, name, value)

    @staticmethod
    def get_model_permissions(model):
        default_permissions = ['{perm}_{object_name}'.format(perm=p, object_name=model._meta.object_name.lower())
                               for p in model._meta.default_permissions]
        return sorted(set(itertools.chain(map(operator.itemgetter(0), model._meta.permissions), default_permissions)))


class ModelNodeMeta(NodeBase):
    """
    Meta class for ``ModelNode``.
    """
    def __new__(mcs, name, bases, attrs):
        cls = super(ModelNodeMeta, mcs).__new__(mcs, str(name), bases, attrs)

        # Set class name and label for node
        cls.__name__ = cls.__label__ = '{object_name}Node'.format(object_name=cls.Meta.model._meta.object_name)

        # Add some default fields
        cls.type = StringProperty(default='ModelNode')
        cls.pk = cls.get_property_class_for_field(cls._pk_field.__class__)(unique_index=True)
        cls._app_label = StringProperty(default=cls.Meta.model._meta.app_label)
        cls._model_name = StringProperty(default=cls.Meta.model._meta.model_name)

        forward_relations = cls.get_forward_relation_fields()
        reverse_relations = cls.get_reverse_relation_fields()

        # Add to cache before recursively looking up relationships.
        from chemtrails.neoutils import __node_cache__
        if cls.Meta.model not in __node_cache__:
            __node_cache__.update({cls.Meta.model: cls})

        for field in cls.Meta.model._meta.get_fields():

            if field.__class__ not in field_property_map:
                continue

            # Add forward relations
            if field in forward_relations:
                relation = cls.get_related_node_property_for_field(field)
                if relation:
                    cls.add_to_class(field.name, relation)
                    if isinstance(field, models.ForeignKey):
                        cls.add_to_class(field.attname,
                                         cls.get_property_class_for_field(field.target_field.__class__)())

            # Add reverse relations
            elif field in reverse_relations:
                relation = cls.get_related_node_property_for_field(field)
                if relation:
                    cls.add_to_class(field.related_name or '%s_set' % field.name, relation)

            # Add concrete fields
            else:
                if field is not cls._pk_field:
                    cls.add_to_class(field.name, cls.get_property_class_for_field(field.__class__)())

        # Recalculate definitions
        cls.__all_properties__ = tuple(cls.defined_properties(aliases=False, rels=False).items())
        cls.__all_aliases__ = tuple(cls.defined_properties(properties=False, rels=False).items())
        cls.__all_relationships__ = tuple(cls.defined_properties(aliases=False, properties=False).items())

        return cls


class ModelNodeMixinBase:
    """
    Base mixin class
    """
    def __update_raw_node__(self):
        """
        Compares the node class attributes to raw node attributes and removes
        any artifacts from the raw node. We want the defined node class 
        attributes to be the source of truth at all times.
        :returns None
        """
        # We can only operate on bound nodes
        if not self._is_bound:
            return

        query = ' '.join((
            'MATCH (n:{label}) WHERE ID(n) = {id}'.format(label=self.__label__, id=self.id),
            'OPTIONAL MATCH (n)-[r]->() RETURN n, r'
        ))
        results, _ = db.cypher_query(query)

        # Identify all properties and relationships which exists on the raw node,
        # but which are not defined on the class model.
        to_remove = defaultdict(set)
        for node, relation in results:

            if not node or node.id in to_remove:
                continue

            for prop in node.properties.keys():
                if prop not in self.defined_properties(aliases=False, rels=False):
                    to_remove[node.id].add(prop)

            relationships = [r.definition['relation_type'] for r in
                             self.defined_properties(aliases=False, properties=False).values()]
            if relation is not None and relation.type not in relationships:
                query = 'MATCH (n:{label})-[r]-() WHERE ID(n) = {id} AND ID(r) = {rid} DELETE r'.format(**{
                    'label': self.__label__,
                    'id': node.id,
                    'rid': relation.id
                })
                db.cypher_query(query)
                logger.info('Relationship type %(type)s is not defined on %(klass)s. '
                            'Removed relationship from raw node.' % {
                                'type': relation.type,
                                'klass': self.__class__.__name__
                            })

            if node.id in to_remove:
                query = ' '.join((
                    'MATCH (n:{label}) WHERE ID(n) = {id} REMOVE'.format(label=self.__label__, id=node.id),
                    ', '.join(['n.%s' % prop for prop in to_remove[node.id]])
                ))
                db.cypher_query(query)
                logger.info('The following %(text)s %(props)s is not defined on %(klass)s. '
                            'Removed %(props)s from raw node.' % {
                                'text': ungettext('property', 'properties', len(to_remove[node.id])),
                                'props': ', '.join([prop for prop in to_remove[node.id]]),
                                'klass': self.__class__.__name__
                            })

    @classproperty
    def _pk_field(cls):
        model = cls.Meta.model
        pk_field = reduce(operator.eq,
                          filter(lambda field: field.primary_key, model._meta.fields))
        return pk_field

    @classproperty
    def _is_ignored(cls):
        lookups = (
            cls.Meta.app_label,
            '{app_label}.*'.format(app_label=cls.Meta.app_label),
            get_model_string(cls.Meta.model)
        )
        from chemtrails.conf import settings
        return any(match in map(str.lower, settings.IGNORE_MODELS) for match in lookups)

    @classproperty
    def has_relations(cls):
        return len(cls.__all_relationships__) > 0

    @property
    def paths(self):
        """
        Returns a ``PathManager`` object which can be used to build 
        traversal paths originating from this node.
        """
        from chemtrails.neoutils.managers import PathManager
        return PathManager(self)

    @staticmethod
    def _log_relationship_definition(action: str, node, prop, level: int = 20):
        direction = {-1: 'INCOMING', 0: 'MUTUAL', 1: 'OUTGOING'}
        message = ('%(action)s %(direction)s relation %(relation_type)s '
                   'between %(klass)r and %(node)r' % {
                       'action': action,
                       'direction': direction[prop.definition['direction']],
                       'relation_type': prop.definition['relation_type'],
                       'klass': prop.source,
                       'node': node
                   })
        logger.log(level=level, msg=message)

    @staticmethod
    def _get_remote_field_name(field):
        reverse_field = True if isinstance(field, (
            models.ManyToManyRel, models.ManyToOneRel, models.OneToOneRel, GenericRelation)) else False

        return str('{model}.{field}'.format(
            model=get_model_string(field.model),
            field=(field.related_name or '%s_set' % field.name
                   if not isinstance(field, (models.OneToOneRel, GenericRelation)) else field.name)) if reverse_field
                   else field.remote_field.field if not isinstance(field, GenericForeignKey)
                   else str(field)).lower()

    @staticmethod
    def get_property_class_for_field(klass):
        """
        Returns the appropriate property class for field class.

        :param klass: Field class which to look up Property type for.
        """
        if klass in field_property_map:
            return field_property_map[klass]
        else:
            # If ``klass`` not found in the field mapping, inspect it's bases
            # and return the first matching base class.
            # NOTE: This might not be very safe, so perhaps we should find a better way.
            for base in klass.__bases__:
                if base in field_property_map:
                    return field_property_map[base]

            # If we're reaching this point we've got a field with an unknown
            # base class. Final attempt at identifying the class type.
            if issubclass(klass, RangeField) and hasattr(klass, 'base_field'):
                # Looks like a PostgreSQL RangeField
                if klass.base_field in field_property_map:
                    return field_property_map[klass.base_field]

        raise NotImplementedError('Unsupported field. Field %s is currently not supported.' % klass.__name__)

    @staticmethod
    def get_cardinality_for_field(field):
        """
        Returns the cardinality for ``field``.
        """
        cardinalities = {
            models.ForeignKey: ZeroOrOne,
            models.ManyToOneRel: ZeroOrMore,
            models.OneToOneField: ZeroOrOne,
            models.OneToOneRel: ZeroOrOne,
            models.ManyToManyField: ZeroOrMore,
            models.ManyToManyRel: ZeroOrMore,
            GenericForeignKey: ZeroOrOne,
            GenericRelation: ZeroOrMore
        }
        return cardinalities[field.__class__]

    @classmethod
    def get_forward_relation_fields(cls):
        return [
            field for field in cls.Meta.model._meta.get_fields()
            if field.is_relation and (
                not field.auto_created or field.concrete
                or field.one_to_one
                or (field.many_to_one and field.related_model)
                or isinstance(field, GenericForeignKey)
            )
        ]

    @classmethod
    def get_reverse_relation_fields(cls):
        return [
            field for field in cls.Meta.model._meta.get_fields()
            if field.auto_created and not field.concrete and (
                field.one_to_many
                or field.many_to_many
                or field.one_to_one
            )
        ]

    @classmethod
    def get_relationship_type(cls, field):
        """
        Get the relationship type for field. If ``settings.NAMED_RELATIONSHIPS``
        is True, pick the relationship type from the field. If not,
        use a pre-defined set of generic types.
        """
        generic_types = {
            Relationship: 'MUTUAL_RELATION',
            RelationshipTo: 'RELATES_TO',
            RelationshipFrom: 'RELATES_FROM'
        }
        if settings.NAMED_RELATIONSHIPS:
            return '{related_name}'.format(
                related_name=getattr(field, 'related_name', field.name) or field.name).upper()
        else:
            prop = cls.get_property_class_for_field(field.__class__)
            return generic_types[prop]

    @classmethod
    def get_related_node_property_for_field(cls, field, meta_node=False):
        """
        Get the relationship definition for the related node based on field.

        :param field: Field to inspect
        :param meta_node: If True, return the meta node for the related model,
                          else return the model node.
        :returns: A ``RelationshipDefinition`` instance.
        """
        from chemtrails.neoutils import (
            __meta_cache__, __node_cache__,
            get_node_class_for_model, get_meta_node_class_for_model
        )

        reverse_field = True if isinstance(field, (
            models.ManyToManyRel, models.ManyToOneRel, models.OneToOneRel, GenericRelation)) else False

        class DynamicRelation(StructuredRel):
            type = StringProperty(default=field.__class__.__name__)
            is_meta = BooleanProperty(default=meta_node)
            remote_field = StringProperty(default=cls._get_remote_field_name(field))
            target_field = StringProperty(default=str(getattr(field, 'target_field', '')).lower())  # NOTE: Workaround for #27

        prop = cls.get_property_class_for_field(field.__class__)
        relationship_type = cls.get_relationship_type(field)

        related_model = field.model if isinstance(field, GenericForeignKey) else field.related_model
        if meta_node:
            klass = (__meta_cache__[related_model]
                     if related_model in __meta_cache__
                     else get_meta_node_class_for_model(related_model))
            return (prop(cls_name=klass, rel_type=relationship_type, model=DynamicRelation)
                         # FIXME: Support cardinality for MetaNodes
                         if not klass._is_ignored else None)
        else:
            klass = (__node_cache__[related_model]
                     if reverse_field and related_model in __node_cache__
                     else get_node_class_for_model(related_model))
            return (prop(cls_name=klass, rel_type=relationship_type, model=DynamicRelation,
                         cardinality=cls.get_cardinality_for_field(field)) if not klass._is_ignored else None)

    @classmethod
    def inflate(cls, node):
        inflated = super(ModelNodeMixinBase, cls).inflate(node)

        model = cls.Meta.model
        app_label = (getattr(inflated._app_label, 'default') if isinstance(inflated._app_label, Property)
                     else inflated._app_label)
        model_name = (getattr(inflated._model_name, 'default') if isinstance(inflated._model_name, Property)
                      else inflated._model_name)

        if app_label != model._meta.app_label:
            raise InflateError(key='_app_label', cls=cls, obj=node, msg=(
                "Cannot inflate node to class %(klass)r. Wrong app label, expected '%(expected)s', "
                "but got '%(actual)s'." % {'klass': cls,
                                           'expected': model._meta.app_label,
                                           'actual': app_label
                                           }))
        elif model_name != model._meta.model_name:
            raise InflateError(key='_model_name', cls=cls, obj=node, msg=(
                "Cannot inflate node to class %(klass)r. Wrong model name, expected '%(expected)s', "
                "but got '%(actual)s'." % {'klass': cls,
                                           'expected': model._meta.model_name,
                                           'actual': model_name
                                           }))
        return inflated

    def sync(self, *args, **kwargs):
        """
        Responsible for syncing the node class with Neo4j.
        """
        raise NotImplementedError('This method must be implemented in derived class!')


class ModelNodeMixin(ModelNodeMixinBase):

    def __init__(self, instance=None, bind=False, *args, **kwargs):
        self._instance = instance
        self.__recursion_depth__ = 0

        defaults = {key: getattr(self._instance, key, kwargs.get(key, None))
                    for key, _ in self.__all_properties__}
        kwargs.update(defaults)
        super(ModelNodeMixinBase, self).__init__(self, *args, **kwargs)

        # Never try to bind ignore nodes
        if bind and self._is_ignored:
            bind = False

        # Query the database for an existing node and set the id if found.
        # This will make this a "bound" node.
        if bind and not hasattr(self, 'id'):
            required = {key: prop for key, prop in self.defined_properties(aliases=False, rels=False).items()
                        if prop.required or prop.unique_index}
            if all(getattr(self, attr) for attr in required.keys()):
                node_id = self._get_id_from_database(self.deflate(self.__properties__))
                if node_id:
                    self.id = node_id

            if not self._instance and self.pk:
                self._instance = self.get_object(self.pk)

        if self._instance:
            # For GenericForeignKey fields, we have no way of knowing what kind of
            # object it is related to during class creation, thus the relationship
            # definition is set to refer itself.
            # When initializing we have access to the underlying django object instance and
            # can inspect it, so we need to switch out the relationship definition
            # node class in order to connect to the correct node.
            from chemtrails.neoutils import get_node_class_for_model
            generic_fks = filter(lambda f: isinstance(f, GenericForeignKey),
                                 self.get_forward_relation_fields())
            for field in generic_fks:
                relationship = self.defined_properties(aliases=False, properties=False).get(field.name, None)
                if isinstance(getattr(self._instance, field.name, None), models.Model) and relationship:
                    node_class = get_node_class_for_model(getattr(self._instance, field.name))
                    model = relationship.definition['model']
                    model.target_field.default = str(node_class._pk_field).lower()

                    # defensively set the definition dictionary to a dict like class
                    # which will protect the safeguarded keys against being copied over
                    # by neomodel
                    relationship.definition = DefinitionDict(relationship.definition)

                    relationship.definition._safeguard('node_class', node_class)
                    relationship.definition._safeguard('model', model)

    def __repr__(self):
        return '<{label}: {id}>'.format(label=self.__class__.__label__,
                                        id=self.id if self._is_bound else None)

    def post_save(self):
        # once a node has been saved, fill in the definition for the attributes
        # that has been protected in the relations. this has to be done in
        # post_save as neomodel throws an exception if it's done before a node is
        # saved
        for attr, relation in self.defined_properties(aliases=False, properties=False).items():
            if relation and hasattr(self, attr) and type(relation.definition) == DefinitionDict:
                prop = getattr(self, attr)

                prop.definition = DefinitionDict(prop.definition)
                prop.definition._safeguard('node_class', relation.definition['node_class'])
                prop.definition._safeguard('model', relation.definition['model'])

    def to_csv(self, cntr=0, target_file=None):
        from chemtrails.neoutils import get_node_for_object
        import csv

        def format_prop(prop_list):
            formated_string = ''
            first = True
            for name, value in prop_list.items():
                if hasattr(value, 'default') and isinstance(value, Property):

                    attr = getattr(self, name, None)

                    if attr is None and value.has_default is False:
                        continue

                    if first:
                        first = False
                    else:
                        formated_string += ', '

                    formated_string += name + ': '

                    v = value.deflate(value=attr)

                    if value.default is None:
                        if isinstance(v, str):
                            formated_string += '\'{value}\''.format(
                                value=v.replace('"', r'\"').replace("'", r"\'"))
                        else:
                            formated_string += str(v)
                    elif isinstance(value, StringProperty):
                        formated_string += '\'{value}\''.format(
                            value=value.default.replace('"', r'\"').replace("'", r"\'"))
                    else:
                        formated_string += str(value.default)

            return formated_string

        prop = self.defined_properties(aliases=False, rels=False)

        source_node_name = self.__label__
        source_node_pk = self.pk

        with open(target_file.name, 'a', newline='') as csvfile:
            spamwriter = csv.writer(csvfile, delimiter=';',
                                    quotechar='|', quoting=csv.QUOTE_MINIMAL)

            queryparams = format_prop(prop)
            spamwriter.writerow(['n'] + ['CREATE (:%s { %s })' % (self.__label__, queryparams)])

            rels = self.defined_properties(aliases=False, properties=False)

            for name, obj in rels.items():

                relation_type = obj.definition['relation_type']
                obj_attr = getattr(self._instance, name, None)
                if not obj_attr:
                    continue

                if isinstance(obj_attr, models.Model):

                    node = get_node_for_object(obj_attr, False)
                    target_node_name = node.__label__
                    target_node_pk = node.pk
                    label_a = source_node_name + str(cntr) + str(source_node_pk)
                    label_b = target_node_name + str(cntr) + str(target_node_pk)
                    label_a = label_a.replace('-', '_')
                    label_b = label_b.replace('-', '_')

                    rel_props = format_prop(obj.definition['model'].__dict__)

                    spamwriter.writerow(
                        ['r'] + [('MATCH({node_a} {{ pk: {node_a_pk} }}), ({node_b} {{ pk: {node_b_pk} }}) '
                                  'WITH {node_a_label}, {node_b_label} '
                                  'CREATE UNIQUE({node_a_label})-[r:{rel}{{{rel_prop}}}]->({node_b_label})'
                                  .format(**{'node_a': label_a + ':' + source_node_name,
                                             'node_a_pk': '"{pk}"'.format(pk=source_node_pk) if isinstance(source_node_pk, str) else source_node_pk,
                                             'node_a_label': label_a,
                                             'node_b': label_b + ':' + target_node_name,
                                             'node_b_pk': '"{pk}"'.format(pk=target_node_pk) if isinstance(target_node_pk, str) else target_node_pk,
                                             'node_b_label': label_b,
                                             'rel': relation_type,
                                             'rel_prop': rel_props
                                             })
                                  )
                                 ]
                    )
                    cntr = cntr + 1

                elif isinstance(obj_attr, Manager):
                    for item in obj_attr.all():
                        i = item

                        node = get_node_for_object(i, False)
                        target_node_name = node.__label__
                        target_node_pk = node.pk
                        label_a = 'a' + str(cntr)
                        label_b = 'b' + str(cntr)
                        label_a = label_a.replace('-', '_')
                        label_b = label_b.replace('-', '_')
                        rel_props = format_prop(obj.definition['model'].__dict__)

                        spamwriter.writerow(
                            ['r'] + [('MATCH({node_a} {{ pk: {node_a_pk} }}), ({node_b} {{ pk: {node_b_pk} }}) '
                                      'WITH {node_a_label}, {node_b_label} '
                                      'CREATE UNIQUE({node_a_label})-[r:{rel}{{{rel_prop}}}]->({node_b_label})'
                                      .format(**{'node_a': label_a + ':' + source_node_name,
                                                 'node_a_pk': '"{pk}"'.format(pk=source_node_pk) if isinstance(source_node_pk, str) else source_node_pk,
                                                 'node_a_label': label_a,
                                                 'node_b': label_b + ':' + target_node_name,
                                                 'node_b_pk': '"{pk}"'.format(pk=target_node_pk) if isinstance(target_node_pk, str) else target_node_pk,
                                                 'node_b_label': label_b,
                                                 'rel': relation_type,
                                                 'rel_prop': rel_props
                                                 })
                                      )
                                     ]
                        )
                        cntr = cntr + 1

    @property
    def _is_bound(self):
        return getattr(self, 'id', None) is not None

    @property
    def _recursion_depth(self):
        return self.__recursion_depth__

    @_recursion_depth.setter
    def _recursion_depth(self, n):
        self.__recursion_depth__ = n

    def _get_id_from_database(self, params):
        """
        Query for node and return id.

        :param params: Parameters to use in query.
        :returns: Node id if found, else None
        """
        query = ' '.join(('MATCH (n:{label}) WHERE'.format(label=self.__label__),
                          ' AND '.join(['n.{} = {{{}}}'.format(key, key) for key in params.keys()]),
                          'RETURN id(n) LIMIT 1'))
        result, _ = db.cypher_query(query, params)
        return list(flatten(result))[0] if result else None

    def get_object(self, pk=None):
        """
        :returns: Django model instance if found or None
        """
        try:
            return self._instance or ContentType.objects.get(
                app_label=self._app_label, model=self._model_name).get_object_for_this_type(
                pk=pk or getattr(self, 'pk', None))
        except ObjectDoesNotExist:
            return None

    def full_clean(self, exclude=None, validate_unique=True):
        exclude = exclude or []

        props = self.__properties__
        for key in exclude:
            if key in props:
                del props[key]

        try:
            self.deflate(props, self)

            if validate_unique:
                cls = self.__class__
                unique_props = [attr for attr, prop in self.defined_properties(aliases=False, rels=False).items()
                                if prop not in exclude and prop.unique_index]
                for key in unique_props:
                    value = self.deflate({key: props[key]}, self)[key]
                    node = cls.nodes.get_or_none(**{key: value})

                    # If exists and not this node
                    if node and node.id != getattr(self, 'id', None):
                        raise ValidationError({key, 'already exists'})

        except DeflateError as e:
            raise ValidationError({e.property_name: e.msg})
        except RequiredProperty as e:
            raise ValidationError({e.property_name: 'is required'})

    @timeit
    def recursive_connect(self, max_depth=settings.MAX_CONNECTION_DEPTH):
        """
        Recursively connect a node branch.

        :param max_depth: Go n nodes deep originating from the current node.
        :returns: None
        """
        from chemtrails.neoutils import get_node_for_object

        def connect(node, prop):
            """
            Connect both sides of the relationship.
            """
            if not prop.is_connected(node) and isinstance(node, prop.definition['node_class']):
                prop.connect(node)
                self._log_relationship_definition('Connected', node, prop)
                for p, r in node.defined_properties(aliases=False, properties=False).items():
                    # Iterate all relationship properties for node and connect the
                    # "reverse" side of the relationship back.
                    # NOTE: Simplify this anyone?
                    p = getattr(node, p)
                    if issubclass(p.definition['node_class'], self.__class__):
                        remote_field = p.definition['model'].remote_field
                        target_field = p.definition['model'].target_field

                        reverse_fields = p.source_class.get_reverse_relation_fields()
                        forward_fields = p.source_class.get_forward_relation_fields()

                        for f in itertools.chain(reverse_fields, forward_fields):
                            if f.__class__ not in field_property_map:
                                continue
                            if (remote_field.default == p.source_class._get_remote_field_name(f)
                                    and target_field.default == str(getattr(f, 'target_field', '')).lower()):

                                instance = self.get_object(self.pk)
                                source = getattr(node.get_object(node.pk), p.name, None)

                                # If `source` is a manager, query for existence of `instance`
                                if ((isinstance(source, Manager) and source.filter(pk=instance.pk).exists())

                                    # Or, if `source` is a model instance, check if it has an attribute
                                    # `p.name` which is the `instance`
                                    or (isinstance(source, models.Model)
                                        and getattr(source, p.name, None) == instance)

                                    # Or, if field `f` is a OneToOneField or OneToOneRel and
                                    # `source` equals the `instance`
                                    or (isinstance(source, models.Model)
                                        and isinstance(f, (models.OneToOneField, models.OneToOneRel))
                                        and source == instance)):

                                    if not p.is_connected(self):
                                        p.connect(self)
                                        self._log_relationship_definition('Connected', self, p)

        def disconnect(node, prop):
            """
            Disconnect both sides of the relationship.
            """
            prop.disconnect(node)
            self._log_relationship_definition('Disconnected', node, prop)
            for p, r in node.defined_properties(aliases=False, properties=False).items():
                p = getattr(node, p)
                if issubclass(p.definition['node_class'], self.__class__):
                    # Iterate all relationship properties for node and disconnect the
                    # "reverse" side of the relationship.
                    remote_field = p.definition['model'].remote_field
                    target_field = p.definition['model'].target_field

                    reverse_fields = p.source_class.get_reverse_relation_fields()
                    forward_fields = p.source_class.get_forward_relation_fields()

                    for f in itertools.chain(reverse_fields, forward_fields):
                        if f.__class__ not in field_property_map:
                            continue
                        if (remote_field.default == p.source_class._get_remote_field_name(f)
                                and target_field.default == str(getattr(f, 'target_field', '')).lower()):
                            p.disconnect(self)
                            self._log_relationship_definition('Disconnected', self, p)

        if max_depth <= 0:
            logger.debug('Reached MAX DEPTH for %(node)r, returning...' % {'node': self})
            return

        # We require a model instance to look for filter values.
        instance = self.get_object(self.pk)
        if not instance:
            return

        logger.info('Mapping relationships for object %(instance)r' % {'instance': instance})
        for attr, relation in self.defined_properties(aliases=False, properties=False).items():
            prop = getattr(self, attr)
            klass = relation.definition['node_class']

            # Never connect ignored models
            if klass._is_ignored:
                continue

            source = getattr(instance, prop.name, None)
            if not source:
                for node in prop.all():
                    disconnect(node, prop)
                continue

            if isinstance(source, models.Model):
                node = klass.nodes.get_or_none(pk=source.pk)
                if not node:
                    node = get_node_for_object(source, bind=False).save()
                    logger.info('Created missing node %(node)r while synchronizing %(instance)r' % {
                        'node': node,
                        'instance': instance
                    })
                if isinstance(prop, ZeroOrOne):
                    # Disconnect previously related node if any
                    to_disconnect = prop.filter(pk__ne=source.pk)
                    for n in to_disconnect:
                        disconnect(n, prop)
                connect(node, prop)
                node.recursive_connect(max_depth=max_depth - 1)

            elif isinstance(source, Manager):
                if not source.exists():
                    to_disconnect = prop.filter(
                        pk__in=list(source.model._meta.default_manager.exclude(pk__in=source.values('pk'))
                                    .values_list('pk', flat=True)))
                    for n in to_disconnect:
                        disconnect(n, prop)
                    continue

                nodeset = klass.nodes.filter(pk__in=list(source.values_list('pk', flat=True)))
                if len(nodeset) != source.count():
                    existing = [n.pk for n in nodeset]
                    for obj in source.exclude(pk__in=existing):
                        node = get_node_for_object(obj, bind=False)
                        if not node._is_ignored:
                            node.save()
                            logger.info('Created missing node %(node)r while synchronizing %(instance)r' % {
                                'node': node,
                                'instance': instance
                            })
                    nodeset = klass.nodes.filter(pk__in=list(source.values_list('pk', flat=True)))

                for node in nodeset:
                    connect(node, prop)
                    node.recursive_connect(max_depth=max_depth - 1)

    def sync(self, max_depth=settings.MAX_CONNECTION_DEPTH, update_existing=True, create_empty=False):
        """
        Synchronizes the current node with data from the database and
        connect all directly related nodes.

        :param max_depth: Maximum depth of recursive connections to be made.
        :param update_existing: If True, save data from the django model to graph node.
        :param create_empty: If the Node has no relational fields, don't create it.
        :returns: The ``ModelNode`` instance or None if not created.
        """
        cls = self.__class__

        if self._is_ignored:
            try:
                self.delete()
            except ValueError:
                pass
            finally:
                return None

        if not cls.has_relations and not create_empty:
            return None

        self.full_clean(validate_unique=not update_existing)

        if update_existing:
            if not self._is_bound:
                node = cls.nodes.get_or_none(**{'pk': self.pk})
                if node:
                    self.id = node.id
            if self._instance is not None:
                # Update the node instance with data from object instance
                # whenever we're syncing.
                defaults = {key: getattr(self._instance, key, None)
                            for key, _ in self.__all_properties__ if hasattr(self._instance, key)}
                for key, value in defaults.items():
                    setattr(self, key, value)

            # Make sure the neo4j node properties and relationships matches
            # the class definition.
            self.__update_raw_node__()

            # Finally save the node.
            self.save()

        # Connect relations
        self.recursive_connect(max_depth=max_depth)

        return self


class MetaNodeMeta(NodeBase):
    """
    Meta class for ``MetaNode``.
    """

    def __new__(mcs, name, bases, attrs):
        cls = super(MetaNodeMeta, mcs).__new__(mcs, str(name), bases, attrs)

        # Set class name and label for node
        cls.__name__ = cls.__label__ = '{object_name}'.format(object_name=cls.Meta.model._meta.object_name)

        # Add some default fields
        cls.type = StringProperty(default='MetaNode')
        cls.label = StringProperty(default=cls.Meta.model._meta.label, unique_index=True)
        cls._app_label = StringProperty(default=cls.Meta.model._meta.app_label)
        cls._model_name = StringProperty(default=cls.Meta.model._meta.model_name)
        cls.model_permissions = ArrayProperty(default=cls.get_model_permissions(cls.Meta.model))
        cls.is_intermediary = BooleanProperty(default=not ContentType.objects.filter(
            app_label=cls.Meta.model._meta.app_label, model=cls.Meta.model._meta.model_name).exists())

        forward_relations = cls.get_forward_relation_fields()
        reverse_relations = cls.get_reverse_relation_fields()

        # Add to cache before recursively looking up relationships.
        from chemtrails.neoutils import __meta_cache__
        if cls.Meta.model not in __meta_cache__:
            __meta_cache__.update({cls.Meta.model: cls})

        # # Add relations for the model
        for field in itertools.chain(forward_relations, reverse_relations):

            if field.__class__ not in field_property_map:
                continue

            # Add forward relations
            if field in forward_relations:
                relation = cls.get_related_node_property_for_field(field, meta_node=True)
                if relation:
                    cls.add_to_class(field.name, relation)

                if settings.CONNECT_META_NODES:
                    relation = cls.get_related_node_property_for_field(field, meta_node=False)
                    if relation:
                        cls.add_to_class('_%s' % field.name, relation)

            # Add reverse relations
            elif field in reverse_relations:
                related_name = field.related_name or '%s_set' % field.name
                relation = cls.get_related_node_property_for_field(field, meta_node=True)
                if relation:
                    cls.add_to_class(related_name, relation)

                if settings.CONNECT_META_NODES:
                    relation = cls.get_related_node_property_for_field(field, meta_node=False)
                    if relation:
                        cls.add_to_class('_%s' % related_name, relation)

        # Recalculate definitions
        cls.__all_properties__ = tuple(cls.defined_properties(aliases=False, rels=False).items())
        cls.__all_aliases__ = tuple(cls.defined_properties(properties=False, rels=False).items())
        cls.__all_relationships__ = tuple(cls.defined_properties(aliases=False, properties=False).items())

        return cls


class MetaNodeMixin(ModelNodeMixin):
    """
    Mixin class for ``StructuredNode`` which adds a number of class methods
    in order to calculate relationship fields from a Django model class.
    """
    def __init__(self, *args, **kwargs):
        self._instance = self.__class__.Meta.model
        self.__recursion_depth__ = 0

        defaults = {key: getattr(self._instance._meta, key, kwargs.get(key, None))
                    for key, _ in self.__all_properties__}
        kwargs.update(defaults)
        StructuredNode.__init__(self, *args, **kwargs)

        # FIXME: Don't look up against id if using inflate with a raw query.
        if not hasattr(self, 'id'):
            params = self.deflate(self.__properties__)

            # Remove any attributes we don't want to include in the MATCH query.
            exclude = ('model_permisssions', 'is_intermediary')
            for attr in exclude:
                if attr in params:
                    del params[attr]

            node_id = self._get_id_from_database(params)
            if node_id:
                self.id = node_id

    @timeit
    def recursive_connect(self, prop, relation, max_depth):
        """
        Recursively connect a related branch.

        :param prop: For example a ``ZeroOrMore`` instance.
        :param relation: ``RelationShipDefinition`` instance
        :param max_depth: Maximum depth of recursive connections to be made.
        :returns: None
        """
        from chemtrails.neoutils import get_meta_node_for_model

        @timeit
        def back_connect(n, depth):
            if n._recursion_depth >= depth:
                return
            n._recursion_depth += 1
            for p, r in n.defined_properties(aliases=False, properties=False).items():
                n.recursive_connect(getattr(n, p), r, max_depth=n._recursion_depth - 1)

        relations = prop.all()
        klass = relation.definition['node_class']
        is_meta = relation.definition['model'].is_meta.default_value()
        if is_meta:
            node = get_meta_node_for_model(klass.Meta.model)
            if not node._is_bound:
                node.save()
            if node not in relations:
                prop.connect(node)
                self._log_relationship_definition('Connected', node, prop)
                back_connect(node, max_depth)
        elif not is_meta and settings.CONNECT_META_NODES:
            for node in relation.definition['node_class'].nodes.all():
                if node not in relations:
                    prop.connect(node)
                    self._log_relationship_definition('Connected', node, prop)
                    back_connect(node, max_depth)

    def sync(self, max_depth=1, update_existing=True, create_empty=False):
        """
        Synchronizes the current node with data from the database and
        connect all directly related nodes.

        :param max_depth: Maximum depth of recursive connections to be made.
        :param update_existing: If True, save data from the django model to graph node.
        :param create_empty: If the Node has no relational fields, don't create it.
        :returns: The ``MetaNode`` instance or None if not created.
        """
        cls = self.__class__

        if self._is_ignored:
            try:
                self.delete()
            except ValueError:
                pass
            finally:
                return None

        if not cls.has_relations and not create_empty:
            return None

        if update_existing:
            if not self._is_bound:
                node = cls.nodes.get_or_none(**{'_app_label': self._app_label,
                                                '_model_name': self._model_name})
                if node:
                    self.id = node.id

            # Make sure the neo4j node properties and relationships matches
            # the class definition.
            self.__update_raw_node__()

            # Finally save the node
            self.save()

        # Connect relations
        for prop, relation in self.defined_properties(aliases=False, properties=False).items():
            prop = getattr(self, prop)
            self.recursive_connect(prop, relation, max_depth=max_depth)
        return self



# -*- coding: utf-8 -*-

from django.apps import apps

from neomodel import db

from rest_framework.response import Response
from rest_framework.viewsets import ViewSet, ModelViewSet

from chemtrails.contrib.permissions.models import AccessRule
from chemtrails.contrib.permissions.serializers import AccessRuleSerializer, NodeSerializer
from chemtrails.neoutils import get_meta_node_class_for_model
from chemtrails.utils import flatten


class AccessRuleViewSet(ModelViewSet):
    """
    API View for access rules.
    """
    # authentication_classes = ()
    # permission_classes = ()
    queryset = AccessRule.objects.all()
    serializer_class = AccessRuleSerializer


class MetaGraphView(ViewSet):
    """
    API View for getting the meta graph.
    """
    # authentication_classes = ()
    # permission_classes = ()

    def list(self, request, format=None):
        """
        Returns the meta graph serialized as JSON.
        """
        result, _ = db.cypher_query('MATCH (n {type: "MetaNode", is_intermediary: False}) RETURN n')

        response = []
        for item in list(flatten(result)):
            try:
                if hasattr(item, 'properties') and all(
                                value in item.properties for value in ('app_label', 'model_name')):
                    model = apps.get_model(app_label=item.properties['app_label'],
                                           model_name=item.properties['model_name'])
                    klass = get_meta_node_class_for_model(model)
                    serializer = NodeSerializer(instance=klass.inflate(item), many=False)
                    response.append(serializer.data)
            except LookupError:
                # Might trigger if getting data originating from a different source.
                continue
        return Response(data=response, content_type='application/json')



# -*- coding: utf-8 -*-

from itertools import chain

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q, Count
from django.shortcuts import _get_queryset
from django.utils.encoding import force_text
from django.utils.translation import ngettext_lazy

from neo4j.v1 import Path
from neomodel import db

from chemtrails.contrib.permissions.exceptions import MixedContentTypeError
from chemtrails.neoutils import InflateError, get_node_class_for_model, get_node_for_object
from chemtrails.contrib.permissions.models import AccessRule
from chemtrails.utils import flatten

User = get_user_model()


def get_identity(identity):
    """
    Returns (user_obj, None) or (None, group_obj) tuple depending on
    given ``identity`` argument.
    @:param  identity: ``User`` or ``Group`` object instance. 
    """
    if isinstance(identity, AnonymousUser):
        raise NotImplementedError('Implement support for AnonymousUser, please!')

    if isinstance(identity, User):
        return identity, None
    elif isinstance(identity, Group):
        return None, identity


def get_content_type(obj):
    """
    Returns the content type for ``obj``.
    """
    return ContentType.objects.get_for_model(obj)


def get_perms(user_or_group, model):
    """
    Return permissions for given user/group and model pair, as
    a list of strings.
    """
    checker = GraphPermissionChecker(user_or_group)
    return checker.get_perms(model)


def get_user_perms(user, model):
    """
    Return permissions for given user and model pair, as a 
    list of strings.
    """
    checker = GraphPermissionChecker(user)
    return checker.get_user_perms(model)


def get_access_rules(ctype_source, ctype_target, codenames):
    """
    Retrieve all defined access rules originating from ``ctype_source``
    and targeting ``ctype_target``.

    :param ctype_source: Content type source
    :param ctype_target: Content type target
    :param codenames: Sequence of permission codenames

    :returns: Queryset containing active access rules.
    """
    # To reduce search space, first retrieve all access rules with n or greater permissions count.
    queryset = (AccessRule.objects.prefetch_related('permissions')
                .annotate(count=Count('permissions')).filter(count__gte=len(codenames)))
    queryset = queryset.filter(ctype_source=ctype_source,
                               ctype_target=ctype_target, is_active=True,
                               permissions__codename__in=codenames)
    return queryset


def check_permissions_app_label(permissions):
    """
    Make sure ``permissions`` matches the ``klass`` app label.
    
    :param permissions: Single permission string or a sequence of permission strings.
      Must be in the format "app_label.codename".
    :type permissions: str or sequence.
    
    :raises MixedContentTypeError: If given a sequence of ``permissions`` with different
      app labels.
    :raises MixedContentTypeError: If computed content type for ``permissions``
      and/or ``klass`` clashes.
    :raises ContentType.DoesNotExist: If failed to look up content type for permission.
    
    :returns: Two tuple with ContentType for permissions and a set of permission codenames.
    :rtype: tuple(ContentType, set(str,))
    """
    if isinstance(permissions, str):
        permissions = [permissions]

    ctype = None
    app_label = None
    codenames = set()

    for perm in permissions:
        if '.' in perm:
            _app_label, codename = perm.split('.', 1)
            if app_label is not None and _app_label != app_label:
                raise MixedContentTypeError('Given permissions must have the same app label '
                                            '(%s != %s).' % (app_label, _app_label))
            app_label = _app_label
        else:
            codename = perm

        codenames.add(codename)
        if app_label is not None:
            _ctype = ContentType.objects.get(app_label=app_label,
                                             permission__codename=codename)
            if ctype is not None and ctype != _ctype:
                raise MixedContentTypeError('Calculated content type from permission "%s" %s does '
                                            'not match %r.' % (perm, _ctype, ctype))
            else:
                ctype = _ctype

    return ctype, codenames


def get_users_with_perms(obj, permissions, with_superusers=False, with_group_users=True):
    """
    Returns a queryset of all ``User`` objects which there can be calculated a path from
    the given ``obj``.

    :param obj: model instance.
    :param permissions: Single permission string, or sequence of permissions strings
      that user requires to have.
    :param with_superusers: Default: ``False``. If set to ``True`` result would
      include all superusers.
    :param with_group_users: Default: ``True``. If set to ``False`` result would
      **not** include users which has only group permissions for given ``obj``.

    :raises MixedContentTypeError: If computed content type for ``permissions``
      and/or ``obj`` clashes.

    :returns: Queryset containing ``User`` objects which has ``permissions`` for ``obj``.
    """
    ctype, codenames = check_permissions_app_label(permissions)
    if ctype is None:
        ctype = get_content_type(obj)
        if codenames:
            # Make sure permissions are valid.
            _codenames = set(ctype.permission_set.filter(codename__in=codenames)
                             .values_list('codename', flat=True))
            if not codenames == _codenames:
                message = ngettext_lazy(
                    'Calculated content type from permission "%s" does not match %r.' % (next(iter(codenames)), ctype),
                    'One or more permissions "%s" from calculated content type does not match %r.' %
                    (', '.join(sorted(codenames)), ctype),
                    len(codenames))
                raise MixedContentTypeError(message)

    elif not ctype == get_content_type(obj):
        raise MixedContentTypeError('Calculated content type %r does not match %r.' % (ctype, get_content_type(obj)))

    queryset = _get_queryset(User)

    # If there is no node in the graph for ``obj``, return empty queryset.
    target_node = get_node_class_for_model(obj).nodes.get_or_none(**{'pk': obj.pk})
    if not codenames or not target_node:
        if with_superusers is True:
            return queryset.filter(is_superuser=True)
        return queryset.none()

    ctype_source = get_content_type(User)

    # We need a fake source content type model to use as origin.
    fake_model = ctype_source.model_class()()
    source_node = get_node_for_object(fake_model, bind=False)

    queries = []
    for access_rule in get_access_rules(ctype_source, ctype, codenames):
        manager = source_node.paths
        if access_rule.direction is not None:
            manager.direction = access_rule.direction

        for n, rule_definition in enumerate(access_rule.relation_types_obj):
            relation_type, target_props = zip(*rule_definition.items())
            relation_type, target_props = relation_type[0], target_props[0]

            source_props = {}
            target_props = target_props or {}

            if n == 0 and access_rule.requires_staff:
                source_props.update({'is_staff': True})

            # Make sure the last object in the query is matched to ``obj``.
            if n == len(access_rule.relation_types_obj) - 1:
                target_props['pk'] = target_node.pk

            # FIXME: Workaround for https://github.com/inonit/django-chemtrails/issues/46
            # If using "{source}.<attr>" filters, ignore them!
            target_props = {key: value for key, value in target_props.items()
                            if isinstance(value, str) and not value.startswith('{source}.')}
            manager = manager.add(relation_type, source_props=source_props, target_props=target_props)

        if manager.statement:
            queries.append(manager.get_path())

    q_values = Q()
    if with_superusers is True:
        q_values |= Q(is_superuser=True)

    start_node_class = get_node_class_for_model(queryset.model)
    end_node_class = get_node_class_for_model(obj)

    with db.transaction:
        results = map(db.cypher_query, queries)

    for result in results:
        if result:
            values = set()
            for item in flatten(result):
                if not isinstance(item, Path):  # pragma: no cover
                    continue
                elif (start_node_class.__label__ not in item.start.labels
                      or end_node_class.__label__ not in item.end.labels):
                    continue
                try:
                    start, end = (start_node_class.inflate(item.start),
                                  end_node_class.inflate(item.end))
                    if isinstance(start, start_node_class) and end == target_node:
                        # Make sure the user object has correct permissions
                        instance = start.get_object()
                        global_perms = set(get_perms(instance, obj) if with_group_users
                                           else get_user_perms(instance, obj))
                        if all((code in global_perms for code in codenames)):
                            values.add(item.start.properties['pk'])
                except (KeyError, InflateError, ObjectDoesNotExist):
                    continue
            q_values |= Q(pk__in=values)

    if not q_values:
        return queryset.none()

    return queryset.filter(q_values)


def get_groups_with_perms(obj, attach_perms=False):
    """
    Returns a queryset of ``Group`` objects which there can be calculated a path from
    the given ``obj``.
    """
    raise NotImplementedError


def get_objects_for_user(user, permissions, klass=None, use_groups=True,
                         extra_perms=None, any_perm=False, with_superuser=True):
    """
    Returns a queryset of objects for which there can be calculated a path between
    the ``user`` using one or more access rules with *all* permissions present
    at ``permissions``.
    
    :param user: ``User`` instance for which objects should be returned.
    :param permissions: Single permission string, or sequence of permission
      strings that should be checked.
      If ``klass`` parameter is not given, those should be full permission
      strings rather than only codenames (ie. ``auth.change_user``). If more than
      one permission is present in the sequence, their content type **must** be 
      the same or ``MixedContentTypeError`` would be raised.
    :param klass: May be a ``Model``, ``Manager`` or ``QuerySet`` object. If not
      given, this will be calculated based on passed ``permissions`` strings.
    :param use_groups: If ``True``, include users groups permissions.
      Defaults to ``True``.
    :param extra_perms: Single permission string, or sequence of permission strings
      that should be used as ``global_perms`` base. These permissions will be 
      treated as if the user possesses them.
    :param any_perm: If ``True``, any permission in sequence is accepted. 
      Defaults to ``False``.
    :param with_superuser: If ``True`` and ``user.is_superuser`` is set, returns
      the entire queryset. Otherwise will only return the objects the user has 
      explicit permissions to. Defaults to ``True``.
     
    :raises MixedContentTypeError: If computed content type for ``permissions``
      and/or ``klass`` clashes.
    :raises ValueError: If unable to compute content type for ``permissions``.
    
    :returns: QuerySet containing objects ``user`` has ``permissions`` to.
    """
    # Make sure all permissions checks out!
    ctype, codenames = check_permissions_app_label(permissions)
    if extra_perms:
        extra_ctype, extra_perms = check_permissions_app_label(extra_perms)
        if extra_ctype != ctype:
            raise MixedContentTypeError('Calculated content type from keyword argument `extra_perms` '
                                        '%s does not match %r.' % (extra_ctype, ctype))
    extra_perms = extra_perms or set()

    if ctype is None and klass is not None:
        queryset = _get_queryset(klass)
        ctype = get_content_type(queryset.model)
    elif ctype is not None and klass is None:
        queryset = _get_queryset(ctype.model_class())
    elif klass is None:
        raise ValueError('Could not determine the content type.')
    else:
        queryset = _get_queryset(klass)
        if ctype.model_class() != queryset.model:
            raise MixedContentTypeError('ContentType for given permissions and klass differs.')

    # Superusers have access to all objects.
    if with_superuser and user.is_superuser:
        return queryset

    # We don't support anonymous users.
    if user.is_anonymous:
        return queryset.none()

    # If there is no node in the graph for the user object, return empty queryset.
    source_node = get_node_class_for_model(user).nodes.get_or_none(**{'pk': user.pk})
    if not source_node:
        return queryset.none()

    # Next, get all permissions the user has, either directly set through user permissions
    # or if ``use_groups`` are set, derived from a group membership.
    global_perms = extra_perms | set(get_perms(user, queryset.model) if use_groups
                                     else get_user_perms(user, queryset.model))

    # Check if we requires the user to have *all* permissions or if it is
    # sufficient with any provided.
    if not any_perm and not all((code in global_perms for code in codenames)):
        return queryset.none()
    elif any_perm:
        for code in codenames.copy():
            if code not in global_perms:
                codenames.remove(code)

    # Calculate a PATH query for each rule
    queries = []
    for access_rule in get_access_rules(get_content_type(user), ctype, codenames):
        manager = source_node.paths
        for n, rule_definition in enumerate(access_rule.relation_types_obj):
            relation_type, target_props = zip(*rule_definition.items())
            relation_type, target_props = relation_type[0], target_props[0]

            source_props = {}
            if n == 0 and access_rule.requires_staff:
                source_props.update({'is_staff': True})
            manager = manager.add(relation_type, source_props=source_props, target_props=target_props)

        if manager.statement:
            queries.append(manager.get_path())

    q_values = Q()
    start_node_class = get_node_class_for_model(user)
    end_node_class = get_node_class_for_model(queryset.model)

    with db.transaction:
        results = map(db.cypher_query, queries)

    for result in results:
        if result:
            values = set()
            for item in flatten(result):
                if not isinstance(item, Path):  # pragma: no cover
                    continue
                elif (start_node_class.__label__ not in item.start.labels
                      or end_node_class.__label__ not in item.end.labels):
                    continue
                try:
                    start, end = (start_node_class(user).inflate(item.start),
                                  end_node_class.inflate(item.end))
                    if start == source_node and isinstance(end, end_node_class):
                        values.add(item.end.properties['pk'])
                except (KeyError, InflateError):  # pragma: no cover
                    continue
            q_values |= Q(pk__in=values)

    # If no values in the Q filter, it means we couldn't get a path from the
    # user node to given object in queryset by any evaluated rule.
    # Return an empty queryset.
    if not q_values:
        return queryset.none()

    return queryset.filter(q_values)


def get_objects_for_group(group, perms, klass=None, any_perm=False, accept_global_perms=True):
    """
    Returns a queryset of objects for which there can be calculated a path....
    """
    raise NotImplementedError


class GraphPermissionChecker(object):
    """
    Object permission checker inspired by django-guardian.
    """
    def __init__(self, user_or_group=None):
        self._obj_perms_cache = {}
        self.user, self.group = get_identity(user_or_group)

    def has_perm(self, perm, obj):
        """
        Checks if a user/group has given permission for object.
        """
        if self.user and not self.user.is_active:
            return False
        elif self.user and self.user.is_superuser:
            return True

        # We only care about the codename.
        perm = perm.split('.')[-1]
        has_perm = perm in self.get_perms(obj) and self.is_authorized(perm, obj)
        return has_perm

    def is_authorized(self, perm, obj):
        """
        Checks if user/group is authorized to access given object.
        """
        target_node = get_node_class_for_model(obj).nodes.get_or_none(**{'pk': obj.pk})
        if not target_node:
            return False

        if self.user:
            queryset = get_objects_for_user(self.user, perm, klass=obj._meta.default_manager.filter(pk=obj.pk))
            return obj in queryset
        elif self.group:
            # TODO: Implement `get_objects_for_group`!
            queryset = get_objects_for_group(self.group, perm, klass=obj._meta.default_manager.filter(pk=obj.pk))
            return obj in queryset

    @staticmethod
    def get_local_cache_key(obj):
        """
        Returns cache key for ``_obj_perms_cache`` dict.
        """
        ctype = get_content_type(obj)
        return ctype.id, force_text(obj.pk)

    @staticmethod
    def get_accessrule_queryset(obj):
        ctype = get_content_type(obj)
        return AccessRule.objects.filter(is_active=True, ctype_target=ctype)

    def get_user_filters(self):
        related_name = User.user_permissions.field.related_query_name()
        user_filters = {'%s' % related_name: self.user}
        return user_filters

    def get_user_perms(self, model):
        ctype = get_content_type(model)
        filters = self.get_user_filters()
        return Permission.objects.filter(content_type=ctype, **filters).values_list('codename', flat=True)

    def get_group_filters(self):
        related_name = Group.permissions.field.related_query_name()
        if self.user:
            field_name = '%s__%s' % (
                related_name,
                User.groups.field.related_query_name()
            )
            group_filters = {field_name: self.user}
        else:
            group_filters = {'%s' % related_name: self.group}

        return group_filters

    def get_group_perms(self, model):
        ctype = get_content_type(model)
        filters = self.get_group_filters()
        return Permission.objects.filter(content_type=ctype, **filters).values_list('codename', flat=True)

    def get_perms(self, obj):
        """
        Returns a list of permissions for given ``obj``.
        """
        if self.user and not self.user.is_active:
            return []

        ctype = get_content_type(obj)
        cache_key = self.get_local_cache_key(obj)
        if cache_key not in self._obj_perms_cache:
            if self.user and self.user.is_superuser:
                perms = list(chain(*Permission.objects
                                   .filter(content_type=ctype)
                                   .values_list('codename')))
            elif self.user:
                user_perms = self.get_user_perms(obj)
                group_perms = self.get_group_perms(obj)
                perms = list(set(chain(user_perms, group_perms)))
            else:
                group_filters = self.get_group_filters()
                perms = list(set(chain(*Permission.objects
                                       .filter(content_type=ctype, **group_filters)
                                       .values_list('codename'))))
            self._obj_perms_cache[cache_key] = perms
        return self._obj_perms_cache[cache_key]



# -*- coding: utf-8 -*-

from django.conf import settings

from rest_framework.filters import BaseFilterBackend


class ChemoPermissionsFilter(BaseFilterBackend):
    """
    A filter backend that limits results to those where the 
    requesting user has read object level permissions based on a
    calculated graph path.
    """
    perm_format = '%(app_label)s.view_%(model_name)s'

    def __init__(self):
        assert 'chemtrails.contrib.permissions' in settings.INSTALLED_APPS, (
            'Using %(class)s, but "chemtrails.contrib.permissions" not found in '
            'INSTALLED_APPS.' % {'class': self.__class__.__name__}
        )

    def filter_queryset(self, request, queryset, view):
        from chemtrails.contrib.permissions.utils import get_objects_for_user

        permission = self.perm_format % {
            'app_label': queryset.model._meta.app_label,
            'model_name': queryset.model._meta.model_name
        }
        return get_objects_for_user(request.user, permissions=permission, klass=queryset)



# -*- coding: utf-8 -*-
from autofixture.autofixtures import UserFixture
from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import post_save, m2m_changed
from django.test import TestCase, override_settings
from django.utils import six

from neomodel import *
from neomodel.match import NodeSet

from chemtrails.neoutils import (
    ModelNodeMeta, ModelNodeMixin, MetaNodeMeta, MetaNodeMixin,
    get_meta_node_class_for_model, get_meta_node_for_model,
    get_node_class_for_model, get_node_for_object, get_nodeset_for_queryset
)
from chemtrails.signals.handlers import post_save_handler, m2m_changed_handler
from chemtrails.utils import flatten

from tests.utils import flush_nodes
from tests.testapp.autofixtures import (
    Author, AuthorFixture,
    Book, BookFixture,
    Publisher,
    Store, StoreFixture,
    Tag
)


class NodeUtilsTestCase(TestCase):
    """
    Test various utility functions for dealing with
    model and node instances.
    """
    def test_get_meta_node_class_for_model(self):
        klass = get_meta_node_class_for_model(Book)
        self.assertTrue(issubclass(klass, StructuredNode))

    @flush_nodes()
    def test_get_meta_node_class_for_concrete_model(self):
        book = BookFixture(Book).create_one()
        klass = get_meta_node_class_for_model(book, for_concrete_model=True)
        self.assertTrue(issubclass(klass, StructuredNode))

    def test_get_node_class_for_model(self):
        klass = get_node_class_for_model(Book)
        self.assertTrue(issubclass(klass, StructuredNode))

    @flush_nodes()
    def test_get_node_class_for_model(self):
        book = BookFixture(Book).create_one()
        klass = get_node_class_for_model(book, for_concrete_model=True)
        self.assertTrue(issubclass(klass, StructuredNode))

    @flush_nodes()
    def test_get_meta_node_for_model(self):
        book_meta = get_meta_node_for_model(Book)
        self.assertIsInstance(book_meta, StructuredNode)

    @flush_nodes()
    def test_get_node_for_object(self):
        store = StoreFixture(Store).create_one(commit=True)
        store_node = get_node_for_object(store)
        self.assertIsInstance(store_node, get_node_class_for_model(Store))

    @flush_nodes()
    def test_get_nodeset_for_queryset(self):
        queryset = Store.objects.filter(pk__in=map(lambda n: n.pk,
                                                   StoreFixture(Store).create(count=2, commit=True)))
        nodeset = get_nodeset_for_queryset(queryset)
        self.assertIsInstance(nodeset, NodeSet)
        for node in nodeset:
            self.assertIsInstance(node, get_node_class_for_model(queryset.model))


class ModelNodeTestCase(TestCase):
    """
    Test that we can create ModelNode instances.
    """
    @flush_nodes()
    def test_create_node_class_adds_to_cache(self):

        klass1 = get_node_class_for_model(Book)

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            class Meta:
                model = Book

        klass2 = get_node_class_for_model(Book)
        self.assertEqual(klass1, klass2)

    @flush_nodes()
    def test_create_model_node(self):
        book = BookFixture(Book).create_one()

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            class Meta:
                model = Book

        self.assertTrue(issubclass(ModelNode, StructuredNode))
        self.assertIsInstance(ModelNode(instance=book), StructuredNode)

    @flush_nodes()
    def test_create_model_with_content_types(self):
        permission = Permission.objects.latest('pk')

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            class Meta:
                model = Permission

        self.assertTrue(issubclass(ModelNode, StructuredNode))
        self.assertIsInstance(ModelNode(instance=permission), StructuredNode)

    def test_create_model_node_declaring_model_in_class(self):

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            __metaclass_model__ = Book

            class Meta:
                model = None

        self.assertEqual(ModelNode.Meta.model, Book)

    def test_create_model_node_custom_app_label(self):

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):

            class Meta:
                model = Book
                app_label = 'custom_app_label'

        self.assertEqual(ModelNode.Meta.app_label, 'custom_app_label')

    def test_create_model_node_fails_without_meta_model(self):
        try:
            @six.add_metaclass(ModelNodeMeta)
            class ModelNode(ModelNodeMixin, StructuredNode):

                class Meta:
                    model = None

            self.fail('Did not fail when defining a ModelNode with missing Meta class model.')
        except ValueError as e:
            self.assertEqual(str(e), '%s.Meta.model attribute cannot be None.' % 'ModelNode')

    def test_create_model_node_fails_without_meta_class(self):
        try:
            @six.add_metaclass(ModelNodeMeta)
            class ModelNode(ModelNodeMixin, StructuredNode):
                pass

            self.fail('Did not fail when defining a ModelNode without a Meta class.')
        except ImproperlyConfigured as e:
            self.assertEqual(str(e), '%s must implement a Meta class.' % 'ModelNode')

    @flush_nodes()
    def test_save_existing_node_is_updated(self):
        group = Group.objects.create(name='a group')

        node1 = get_node_for_object(group)
        self.assertEqual(group.name, node1.name)

        group.name = 'still the same group'
        group.save()

        node2 = get_node_for_object(group)
        self.assertEqual(group.name, node2.name)

        self.assertEqual(node1.id, node2.id)

    @flush_nodes()
    def test_sync_existing_node_is_updated(self):
        group = Group.objects.create(name='a group')

        node = get_node_for_object(group)
        self.assertEqual(group.name, node.name)

        group.name = 'another name'
        group.save()

        node.sync()
        self.assertEqual(group.name, node.name)

    @flush_nodes()
    def test_inflate_node_class_raises_inflate_error(self):
        # Make sure the InflateError is raised if trying
        # to inflate nodes with wrong model
        author = AuthorFixture(Author).create_one()
        deflated, _ = list(flatten(db.cypher_query('MATCH (u: UserNode {pk: %d}) RETURN u;' % author.user.pk)))

        node_class = get_node_class_for_model(Author)
        self.assertRaises(InflateError, node_class.inflate, deflated)

    def test_deflate_empty_node_raise_required_property_error(self):
        klass = get_node_class_for_model(User)
        self.assertRaisesMessage(RequiredProperty, "property 'pk' on objects of class UserNode",
                                 klass.deflate, klass().__properties__)


class MetaNodeTestCase(TestCase):
    """
    Test that we can create "singleton" meta nodes
    """
    def test_create_meta_node(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):

            class Meta:
                model = Book

        self.assertTrue(issubclass(MetaNode, StructuredNode))
        self.assertIsInstance(MetaNode(), StructuredNode)

    def test_create_meta_node_declaring_model_in_class(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):
            __metaclass_model__ = Book

            class Meta:
                model = None

        self.assertEqual(MetaNode.Meta.model, Book)

    def test_create_meta_node_custom_app_label(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):

            class Meta:
                model = Book
                app_label = 'custom_app_label'

        self.assertEqual(MetaNode.Meta.app_label, 'custom_app_label')

    def test_create_meta_node_custom_permissions(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):
            class Meta:
                model = Publisher

        self.assertEqual(MetaNode.Meta.model, Publisher)

    def test_create_meta_node_fails_without_meta_model(self):
        try:
            @six.add_metaclass(MetaNodeMeta)
            class MetaNode(MetaNodeMixin, StructuredNode):

                class Meta:
                    model = None

            self.fail('Did not fail when defining a MetaNode with missing Meta class model.')
        except ValueError as e:
            self.assertEqual(str(e), '%s.Meta.model attribute cannot be None.' % 'MetaNode')

    def test_create_meta_node_fails_without_meta_class(self):
        try:
            @six.add_metaclass(MetaNodeMeta)
            class MetaNode(MetaNodeMixin, StructuredNode):
                pass

            self.fail('Did not fail when defining a MetaNode without a Meta class.')
        except ImproperlyConfigured as e:
            self.assertEqual(str(e), '%s must implement a Meta class.' % 'MetaNode')

    @flush_nodes()
    @override_settings(CHEMTRAILS={
        'CONNECT_META_NODES': True
    })
    def test_connected_meta_node(self):
        book = BookFixture(Book).create_one()
        meta = get_meta_node_for_model(Book).sync()

        # FIXME: Settings object is not updated when using override_settings


class GraphMapperTestCase(TestCase):
    """
    Test that node relationships are mapped correctly
    """
    def test_get_property_class_for_field(self):
        from django.contrib.contenttypes.fields import GenericRelation
        from django.contrib.postgres.fields import (
            ArrayField, HStoreField, JSONField,
            IntegerRangeField, BigIntegerRangeField, FloatRangeField,
            DateTimeRangeField, DateRangeField
        )
        from django.db import models

        class CustomField(object):
            pass

        klass = get_node_class_for_model(Book)

        self.assertEqual(klass.get_property_class_for_field(models.ForeignKey), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ForeignKey), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.OneToOneField), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ManyToManyField), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ManyToOneRel), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.OneToOneRel), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ManyToManyRel), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(GenericRelation), RelationshipTo)

        self.assertEqual(klass.get_property_class_for_field(models.AutoField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.BigAutoField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.BooleanField), BooleanProperty)
        self.assertEqual(klass.get_property_class_for_field(models.CharField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.CommaSeparatedIntegerField), ArrayProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DateField), DateProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DateTimeField), DateTimeProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DecimalField), FloatProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DurationField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.EmailField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.FilePathField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.FileField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.FloatField), FloatProperty)
        self.assertEqual(klass.get_property_class_for_field(models.GenericIPAddressField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.IntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.IPAddressField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.NullBooleanField), BooleanProperty)
        self.assertEqual(klass.get_property_class_for_field(models.PositiveIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.PositiveSmallIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.SlugField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.SmallIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.TextField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.TimeField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.URLField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.UUIDField), StringProperty)

        # Test special fields
        # self.assertEqual(klass.get_property_class_for_field(ArrayField), ArrayProperty)
        self.assertEqual(klass.get_property_class_for_field(HStoreField), JSONProperty)
        self.assertEqual(klass.get_property_class_for_field(JSONField), JSONProperty)

        # Test undefined fields by inspecting their base classes.
        self.assertEqual(klass.get_property_class_for_field(models.BigIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(IntegerRangeField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(BigIntegerRangeField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(FloatRangeField), FloatProperty)
        self.assertEqual(klass.get_property_class_for_field(DateTimeRangeField), DateTimeProperty)
        self.assertEqual(klass.get_property_class_for_field(DateRangeField), DateProperty)

        # Test unsupported field
        self.assertRaisesMessage(
            NotImplementedError, 'Unsupported field. Field CustomField is currently not supported.',
            klass.get_property_class_for_field, CustomField)

    @flush_nodes()
    def test_update_raw_node_property(self):
        group = Group.objects.create(name='group')
        node = get_node_for_object(group)

        # Set a custom attribute on the node and make sure it's saved on the node
        result, _ = list(flatten(db.cypher_query('MATCH (n) WHERE ID(n) = %d '
                                                 'SET n.foo = "bar", n.baz = "qux" RETURN n' % node.id)))
        self.assertTrue(all(i in result.properties for i in ('foo', 'baz')))
        self.assertEqual(result.properties['foo'], 'bar')
        self.assertEqual(result.properties['baz'], 'qux')

        self.assertFalse(all(hasattr(node, i) for i in ('foo', 'baz')))
        node.__update_raw_node__()

        # Make sure the custom attribute has been deleted
        result, _ = list(flatten(db.cypher_query('MATCH (n) WHERE ID(n) = %d RETURN n' % node.id)))
        self.assertFalse(all(i in result.properties for i in ('foo', 'baz')))

    @flush_nodes()
    def test_update_raw_node_relationship(self):
        group1, group2 = Group.objects.create(name='group1'), Group.objects.create(name='group2')
        node1, node2 = get_node_for_object(group1), get_node_for_object(group2)

        # Create a custom relationship between group1 and group2
        results, _ = db.cypher_query(
            'MATCH (a), (b) WHERE ID(a) = %d AND ID(b) = %d '
            'CREATE (a)-[r1:RELATION]->(b), (b)-[r2:RELATION]->(a) '
            'RETURN r1, r2' % (node1.id, node2.id)
        )
        results = list(flatten(results))
        self.assertEqual(len(results), 2)
        self.assertTrue(all([r.type == 'RELATION' for r in results]))

        node1.__update_raw_node__()

        # Make sure custom relationship is deleted
        results, _ = db.cypher_query('MATCH (n)-[r]->() WHERE ID(n) = %d RETURN r' % node1.id)
        self.assertEqual(len(results), 0)

        # The other relationship should still be intact
        results, _ = db.cypher_query('MATCH (n)-[r]->() WHERE ID(n) = %d RETURN r' % node2.id)
        results = list(flatten(results))
        self.assertEqual(len(results), 1)

    @override_settings(CHEMTRAILS={
        'IGNORE_MODELS': ['auth']
    })
    def test_ignored_models_app_label(self):
        klass = get_node_class_for_model(Group)
        self.assertTrue(klass._is_ignored)

        klass = get_node_class_for_model(Book)
        self.assertFalse(klass._is_ignored)

    @override_settings(CHEMTRAILS={
        'IGNORE_MODELS': ['auth.*']
    })
    def test_ignored_models_app_label_wildcard(self):
        klass = get_node_class_for_model(Group)
        self.assertTrue(klass._is_ignored)

        klass = get_node_class_for_model(Book)
        self.assertFalse(klass._is_ignored)

    @override_settings(CHEMTRAILS={
        'IGNORE_MODELS': ['auth.group']
    })
    def test_ignored_models_app_label_model_name(self):
        klass = get_node_class_for_model(Group)
        self.assertTrue(klass._is_ignored)

        klass = get_node_class_for_model(Permission)
        self.assertFalse(klass._is_ignored)

    @override_settings(CHEMTRAILS={
        'IGNORE_MODELS': ['auth.Group']
    })
    def test_ignore_models_case_insensitive(self):
        klass = get_node_class_for_model(Group)
        self.assertTrue(klass._is_ignored)

    @flush_nodes()
    @override_settings(CHEMTRAILS={
        'IGNORE_MODELS': ['auth.group']
    })
    def test_ignored_model_sync(self):
        group = Group.objects.create(name='group')

        get_node_for_object(group).sync()
        klass = get_node_class_for_model(Group)
        try:
            klass.nodes.get(pk=group.pk)
            self.fail('Did not fail when trying to look up an ignored node class.')
        except klass.DoesNotExist as e:
            self.assertEqual(str(e), "{'pk': %d}" % group.pk)

    @flush_nodes()
    @override_settings(CHEMTRAILS={
        'IGNORE_MODELS': ['auth.group']
    })
    def test_related_ignore_model_sync(self):
        user = User.objects.create_user(username='testuser', password='test123.')
        group = Group.objects.create(name='group')
        user.groups.add(group)

        get_node_for_object(user).sync()

        klass = get_node_class_for_model(Group)
        try:
            klass.nodes.get(pk=group.pk)
            self.fail('Did not fail when trying to look up an ignored node class.')
        except klass.DoesNotExist as e:
            self.assertEqual(str(e), "{'pk': %d}" % group.pk)

    @flush_nodes()
    def test_change_one_to_one_field_model_sync(self):
        # Make sure we're not getting ``AttemptedCardinalityViolation`` when trying to
        # change a OneToOneField constraint.
        author = AuthorFixture(Author).create_one()
        user = User.objects.create_user(username='testuser', password='test123.')

        author.user = user
        author.save()
        self.assertEqual(author.user, user)

    @flush_nodes()
    def test_recursive_connect(self):
        post_save.disconnect(post_save_handler, dispatch_uid='chemtrails.signals.handlers.post_save_handler')
        m2m_changed.disconnect(m2m_changed_handler, dispatch_uid='chemtrails.signals.handlers.m2m_changed_handler')
        try:
            book = BookFixture(Book, generate_m2m={'authors': (1, 1)}).create_one()
            for depth in range(3):
                db.cypher_query('MATCH (n)-[r]-() WHERE n.type = "ModelNode" DELETE r')  # Delete all relationships
                book_node = get_node_for_object(book).save()
                book_node.recursive_connect(depth)

                if depth == 0:
                    # Max depth 0 means that no recursion should occur, and no connections
                    # can be made, because the connected objects might not exist.
                    for prop in book_node.defined_properties(aliases=False, properties=False).keys():
                        relation = getattr(book_node, prop)
                        try:
                            self.assertEqual(len(relation.all()), 0)
                        except CardinalityViolation:
                            # Will raise CardinalityViolation for nodes which has a single
                            # required relationship
                            continue

                elif depth == 1:
                    self.assertEqual(0, len(get_node_class_for_model(Book).nodes.has(store_set=True)))
                    self.assertEqual(0, len(get_node_class_for_model(Store).nodes.has(books=True)))

                    self.assertEqual(0, len(get_node_class_for_model(Book).nodes.has(bestseller_stores=True)))
                    self.assertEqual(0, len(get_node_class_for_model(Store).nodes.has(bestseller=True)))

                    self.assertEqual(1, len(get_node_class_for_model(Book).nodes.has(publisher=True)))
                    self.assertEqual(1, len(get_node_class_for_model(Publisher).nodes.has(book_set=True)))

                    self.assertEqual(1, len(get_node_class_for_model(Book).nodes.has(authors=True)))
                    self.assertEqual(1, len(get_node_class_for_model(Author).nodes.has(book_set=True)))

                    self.assertEqual(0, len(get_node_class_for_model(Author).nodes.has(user=True)))
                    self.assertEqual(0, len(get_node_class_for_model(User).nodes.has(author=True)))

                    self.assertEqual(1, len(get_node_class_for_model(Book).nodes.has(tags=True)))
                    self.assertEqual(0, len(get_node_class_for_model(Tag).nodes.has(content_type=True)))

                elif depth == 2:
                    self.assertEqual(1, len(get_node_class_for_model(Author).nodes.has(user=True)))
                    self.assertEqual(1, len(get_node_class_for_model(User).nodes.has(author=True)))
                    self.assertEqual(1, len(get_node_class_for_model(Tag).nodes.has(content_type=True)))
                    self.assertEqual(1, len(get_node_class_for_model(ContentType)
                                            .nodes.has(content_type_set_for_tag=True)))
        finally:
            post_save.connect(post_save_handler, dispatch_uid='chemtrails.signals.handlers.post_save_handler')
            m2m_changed.connect(m2m_changed_handler, dispatch_uid='chemtrails.signals.handlers.m2m_changed_handler')



# -*- coding: utf-8 -*-

import json
from collections import defaultdict, OrderedDict

from django.core import exceptions
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.utils.translation import ugettext_lazy as _

from chemtrails.contrib.permissions import forms


class JSONField(models.Field):
    empty_strings_allowed = False
    description = _('An ordered JSON object')
    default_error_messages = {
        'invalid': _("'%(value)s' is not a valid JSON string.")
    }

    def __init__(self, *args, **kwargs):
        """
        :param dump_kwargs: Keyword arguments which will be passed to `json.dumps()`
        :param load_kwargs: Keyword arguments which will be passed to `json.loads()`
        """
        self.dump_kwargs = kwargs.pop('dump_kwargs', {
            'cls': DjangoJSONEncoder,
            'ensure_ascii': False,
            'sort_keys': False,
            'separators': (',', ':')
        })
        self.load_kwargs = kwargs.pop('load_kwargs', {
            'object_pairs_hook': OrderedDict
        })
        if not kwargs.get('null', False):
            kwargs['default'] = kwargs.get('default', defaultdict(OrderedDict))
        super(JSONField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        defaults = {
            'dump_kwargs': self.dump_kwargs,
            'load_kwargs': self.load_kwargs,
            'form_class': forms.JSONFormField,
            'widget': forms.JSONWidget
        }
        return super(JSONField, self).formfield(**defaults)

    def from_db_value(self, value, *args, **kwargs):
        """
        Convert the JSON string to an OrderedDict.
        """
        if value is None:
            return None
        if isinstance(value, (dict, OrderedDict)):
            value = json.dumps(value, **self.dump_kwargs)
        return value

    def get_default(self):
        default = super(JSONField, self).get_default()
        if default and isinstance(default, (dict, OrderedDict)):
            return json.dumps(default, **self.dump_kwargs)
        return default

    def get_db_prep_value(self, value, connection, prepared=False):
        return self.get_prep_value(value)

    def get_internal_type(self):
        return "TextField"

    def get_prep_value(self, value):
        """
        Convert the value to a JSON string ready to
        be stored in the database.
        """
        if value is None:
            if not self.null and self.blank:
                return ''
            return None
        return self.value_to_string(value)

    def to_python(self, value):
        return json.loads(value, **self.load_kwargs)

    def validate(self, value, model_instance):
        if not self.null and value is None:
            raise exceptions.ValidationError(self.error_messages['null'])
        try:
            self.get_prep_value(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(self.error_messages['invalid'],
                                             code='invalid', params={'value': value})

    def value_to_string(self, value):
        try:
            value = self.value_from_object(value)
        except AttributeError:
            pass
        if isinstance(value, (dict, OrderedDict)):
            value = json.dumps(value, **self.dump_kwargs)
        return json.dumps(json.loads(value, **self.load_kwargs), **self.dump_kwargs)



# -*- coding: utf-8 -*-

import os

from django.test import TestCase
from chemtrails import utils


class UtilsTestCase(TestCase):

    def setUp(self):
        os.environ['TRUE'] = 'TRUE'
        os.environ['True'] = 'True'
        os.environ['true'] = 'true'
        os.environ['FALSE'] = 'FALSE'
        os.environ['False'] = 'False'
        os.environ['false'] = 'false'
        os.environ['0'] = '0'
        os.environ['1'] = '1'
        os.environ['1.0'] = '1.0'

    def test_get_environment_variable_bool(self):
        self.assertTrue(utils.get_environment_variable('TRUE'))
        self.assertTrue(utils.get_environment_variable('True'))
        self.assertTrue(utils.get_environment_variable('true'))

        self.assertFalse(utils.get_environment_variable('FALSE'))
        self.assertFalse(utils.get_environment_variable('False'))
        self.assertFalse(utils.get_environment_variable('false'))

    def test_get_environment_variable_int(self):
        self.assertEqual(utils.get_environment_variable('0'), 0)
        self.assertEqual(utils.get_environment_variable('1'), 1)

    def test_get_environment_variable_float(self):
        self.assertEqual(utils.get_environment_variable('1.0'), 1.0)



# -*- coding: utf-8 -*-
import json
from collections import OrderedDict

from django import forms
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from neomodel import config


class GraphWidget(forms.Widget):
    """
    Django admin widget for rendering a Neo4j graph.
    """

    template = 'admin/chemtrails/contrib/permissions/graph_widget.html'

    class Media:
        js = (
            'https://cdnjs.cloudflare.com/ajax/libs/react/15.4.2/react.min.js',
            'chemtrails/contrib/permissions/js/main.js'
        )
        css = {
            'all': (
                'chemtrails/contrib/permissions/css/graph-widget.css',
                'https://cdnjs.cloudflare.com/ajax/libs/semantic-ui/2.2.9/semantic.min.css',)
        }

    def render(self, name, value, attrs=None):
        context = {
            'INITIAL_STATE': json.dumps({
                'settings': {
                    'baseUrl': 'localhost',
                    'neo4jUrl': config.DATABASE_URL
                },
            })
        }
        return mark_safe(
            render_to_string(self.template, context=context)
        )


class JSONWidget(forms.Textarea):
    """
    A simple text widget for displaying JSON strings.
    """
    def render(self, name, value, attrs=None):
        if not value:
            value = ''
        if value and isinstance(value, str):
            value = json.dumps(json.loads(value, object_pairs_hook=OrderedDict),
                               ensure_ascii=False, indent=2, sort_keys=False, separators=(',', ':'))
        return super(JSONWidget, self).render(name, value, attrs)


class CypherWidget(forms.Textarea):
    class Media:
        js = (
            'https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.26.0/codemirror.min.js',
            'https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.26.0/mode/cypher/cypher.min.js',
            'chemtrails/contrib/permissions/js/cypher-widget.js'
        )
        css = {
            'all': (
                'https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.26.0/codemirror.min.css',
                'https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.26.0/theme/neo.min.css'
            )
        }




# -*- coding: utf-8 -*-

from django.contrib.auth.backends import ModelBackend
from django.db import models

from chemtrails.contrib.permissions.utils import GraphPermissionChecker, check_permissions_app_label


def check_user_support(user_obj):
    return not user_obj.is_anonymous and user_obj.is_active


class ChemoPermissionsBackend(ModelBackend):
    """
    Graph based permission backend for Django.
    
    This backend is built around the idea that entities has to have 
    some sort of relationships between them in order to let users 
    access them. It is done by querying Neo4j for a PATH between
    the ``user_obj`` node and the ``obj`` node. One or more ``AccessRule``
    objects are used to build the PATH query and required permissions.
    """
    def authenticate(self, username=None, password=None, **kwargs):
        """
        ``ChemoPermissionsBackend`` does not support authentication.
        """
        return None

    def has_perm(self, user_obj, perm, obj=None):
        """
        Returns ``True`` if given ``user_obj`` has ``perm`` for ``obj``. If
        no ``obj`` is given, return ``False``.
        """
        if not isinstance(obj, models.Model):
            return False

        if not check_user_support(user_obj):
            return False

        if '.' in perm:
            check_permissions_app_label(perm)

        checker = GraphPermissionChecker(user_obj)
        return checker.has_perm(perm, obj)

    def get_all_permissions(self, user_obj, obj=None):
        """
        Returns a list of permission strings that the given ``user_obj`` has for ``obj``.
        """
        if not isinstance(obj, models.Model) or not check_user_support(user_obj):
            return set()

        checker = GraphPermissionChecker(user_obj)
        return checker.get_perms(obj)



# -*- coding: utf-8 -*-

import json

from django import forms
from django.utils.translation import ugettext_lazy as _
from chemtrails.contrib.permissions.forms.widgets import JSONWidget


class JSONFormField(forms.CharField):
    default_error_messages = {
        'invalid': _("'%(value)s' is not a valid JSON string. JSON decode error: %(error)s")
    }

    def __init__(self, **kwargs):
        kwargs.setdefault('widget', JSONWidget)
        self.dump_kwargs = kwargs.pop('dump_kwargs', {})
        self.load_kwargs = kwargs.pop('load_kwargs', {})
        super(JSONFormField, self).__init__(**kwargs)

    def to_python(self, value):
        if self.disabled:
            return value
        if value in self.empty_values:
            return None
        try:
            if value and isinstance(value, str):
                value = json.dumps(json.loads(value, **self.load_kwargs), **self.dump_kwargs)
            return value
        except ValueError as e:
            raise forms.ValidationError(self.error_messages['invalid'],
                                        code='invalid', params={'value': value, 'error': str(e.args[0])})



# -*- coding: utf-8 -*-

from django import forms
from django.conf.urls import url, include
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.text import Truncator

from rest_framework import routers
from rest_framework.decorators import api_view
from rest_framework.response import Response

from chemtrails.contrib.permissions.models import AccessRule
from chemtrails.contrib.permissions.views import AccessRuleViewSet, MetaGraphView
from chemtrails.contrib.permissions.forms import CypherWidget
from chemtrails.neoutils.query import get_node_relationship_types
from chemtrails.neoutils import get_node_for_object
from chemtrails.neoutils.query import validate_cypher


class AccessRuleForm(forms.ModelForm):

    cypher_statement = forms.CharField(widget=CypherWidget, required=False,
                                       help_text=_('Preview the generated cypher statement. Any changes done in '
                                                   'this view will <b>not</b> be saved. <br />Note that this '
                                                   'is generated using a mock source node, which means that filter '
                                                   'parameters will be replaced with real values on evaluation.'
                                                   ''))

    class Meta:
        model = AccessRule
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        kwargs['initial'] = {'cypher_statement': self.get_statement(kwargs.get('instance', None))}
        super(AccessRuleForm, self).__init__(*args, **kwargs)

    def get_statement(self, instance=None):
        if not isinstance(instance, self._meta.model):
            return ''

        # TODO: This should be DRY'ed up!
        fake_model = instance.ctype_source.model_class()(pk=0)
        manager = get_node_for_object(fake_model, bind=False).paths
        if instance.direction is not None:
            manager.direction = instance.direction

        error_message = _('Unable to validate cypher statement.\nError was: "%(error)s".')

        query = None
        for n, rule_definition in enumerate(instance.relation_types_obj):
            relation_type, target_props = zip(*rule_definition.items())
            relation_type, target_props = relation_type[0], target_props[0]

            source_props = {}
            if n == 0 and instance.requires_staff:
                source_props.update({'is_staff': True})
            try:
                manager = manager.add(relation_type, source_props=source_props, target_props=target_props)
            except (ValueError, AttributeError) as e:
                return error_message % {'error': e}

        if manager.statement:
            query = manager.get_path()

        try:
            if query:
                return query
                # FIXME: Bug in libcypher-parser-python
                # https://github.com/inonit/libcypher-parser-python/issues/1
                # validate_cypher(query, raise_exception=True, exc_class=ValidationError)
        except ValidationError as e:
            return error_message % {'error': e}
        return query


class TargetContentTypeFilter(SimpleListFilter):
    title = _('target content type')
    parameter_name = 'ctype_target__model'

    def lookups(self, request, model_admin):
        queryset = model_admin.get_queryset(request)
        return [(value, value) for value in queryset.values_list(self.parameter_name, flat=True)
                .distinct().order_by(self.parameter_name)]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(models.Q(**{'%s__exact' % self.parameter_name: self.value()}))


@admin.register(AccessRule)
class AccessRuleAdmin(admin.ModelAdmin):
    form = AccessRuleForm
    actions = ('toggle_active',)
    list_display = ('short_description', 'ctype_target', 'ctype_source',
                    'requires_staff', 'is_active', 'updated')
    list_filter = ('requires_staff', 'is_active', TargetContentTypeFilter)
    search_fields = ('description',)
    filter_horizontal = ('permissions',)
    fieldsets = (
        (None, {'fields': ('ctype_source', 'ctype_target', 'description', 'permissions',
                           'relation_types', 'direction', 'cypher_statement', 'requires_staff', 'is_active')}),
        ('Dates', {'fields': ('created', 'updated')})
        # ('Rule editor', {'fields': ('graph',)}),
    )
    formfield_overrides = {
        ArrayField: {'widget': forms.Textarea}
    }
    readonly_fields = ('created', 'updated')

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        formfield = super(AccessRuleAdmin, self).formfield_for_foreignkey(db_field, request, **kwargs)

        # Order `ctype_source` and `ctype_target` by model name
        if db_field.name in ('ctype_source', 'ctype_target'):
            formfield.queryset = formfield.queryset.order_by('model')
        return formfield

    def toggle_active(self, request, queryset):
        """
        Inverts the ``is_active`` flag of chosen access rules.
        """
        queryset.update(
            is_active=models.Case(
                models.When(is_active=True, then=models.Value(False)),
                default=models.Value(True)))
        self.message_user(request, _('Activated {0} and deactivated {1} '
                                     'access rules.'.format(queryset.filter(is_active=True).count(),
                                                            queryset.filter(is_active=False).count())))
    toggle_active.short_description = _('Toggle active or inactive access rules')

    def short_description(self, obj):
        return Truncator(obj.description).chars(65)

    def get_urls(self):

        router = routers.DefaultRouter()
        router.register(r'access-rules', AccessRuleViewSet)
        router.register(r'meta-graph', MetaGraphView, base_name='metagraph')

        info = self.model._meta.app_label, self.model._meta.model_name
        urlpatterns = [
            url(r'^neo4j/nodelist/$', self.get_nodelist_api_view, name='%s_%s_nodelist' % info),  # Deprecated
            url(r'^neo4j/', include(router.urls))
        ] + super(AccessRuleAdmin, self).get_urls()
        return urlpatterns

    @staticmethod
    @api_view(http_method_names=['GET'])
    def get_nodelist_api_view(request):
        # TODO: This should be removed!!
        params = {'type': 'MetaNode'}
        result = get_node_relationship_types(params)
        return Response(data=result)



# -*- coding: utf-8 -*-

from django.utils import six

from neomodel import *
from chemtrails.neoutils.core import (
    ModelNodeMeta, ModelNodeMixin,
    MetaNodeMeta, MetaNodeMixin
)

__all__ = [
    'get_meta_node_class_for_model',
    'get_meta_node_for_model',
    'get_node_class_for_model',
    'get_node_for_object',
    'get_nodeset_for_queryset',
    '__meta_cache__',
    '__node_cache__'
]

# Caches to avoid infinity loops
__meta_cache__ = {}
__node_cache__ = {}


def get_meta_node_class_for_model(model, for_concrete_model=True):
    """
    Meta nodes are used to generate a map of the relationships
    in the database. There's only a single MetaNode per model.
    :param model: Django model class.
    :returns: A ``StructuredNode`` class.
    """
    if for_concrete_model:
        model = model._meta.concrete_model

    if model in __meta_cache__:
        return __meta_cache__[model]
    else:
        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):
            __metaclass_model__ = model

            class Meta:
                model = None  # Will pick model from parent class __metaclass_model__ attribute

        __meta_cache__[model] = MetaNode
        return MetaNode


def get_meta_node_for_model(model):
    """
    Get a ``MetaNode`` instance for the current model class.
    :param model: Django model class.
    :returns: A ``MetaNode`` instance.
    """
    klass = get_meta_node_class_for_model(model)
    return klass()


def get_node_class_for_model(model, for_concrete_model=True):
    """
    Model nodes represent a model instance in the database.
    :param model: Django model class.
    :returns: A ``ModelNode`` class.
    """
    if for_concrete_model:
        model = model._meta.concrete_model

    if model in __node_cache__:
        return __node_cache__[model]
    else:
        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            __metaclass_model__ = model

            class Meta:
                model = None  # Will pick model from parent class __metaclass_model__ attribute

        __node_cache__[model] = ModelNode
        return ModelNode


def get_node_for_object(instance, bind=True):
    """
    Get a ``ModelNode`` instance for the current object instance.
    :param instance: Django model instance.
    :param bind: Boolean indicating if we should instantiate a bound node.
    :returns: A ``ModelNode`` instance.
    """
    klass = get_node_class_for_model(instance, True)
    return klass(instance=instance, bind=bind)


def get_nodeset_for_queryset(queryset, sync=False, max_depth=1):
    """
    Get a ``NodeSet`` instance for the current queryset instance.
    :param queryset: Django ``QuerySet`` instance.
    :param sync: Sync all items in the queryset before returning.
    :param max_depth: Maximum depth of recursive connections to be made
                      while syncing each node in the nodeset.
    :returns: A ``neomodel.match.NodeSet`` instance.
    """
    klass = get_node_class_for_model(queryset.model)
    nodeset = klass.nodes.filter(pk__in=list(queryset.values_list('pk', flat=True)))
    if sync:
        for instance in queryset:
            get_node_for_object(instance, bind=False).sync(max_depth=max_depth, update_existing=True)
        nodeset = get_nodeset_for_queryset(queryset, sync=False)
    return nodeset





from django.core.management.base import BaseCommand, CommandError
from chemtrails.neoutils import get_node_class_for_model
from django.apps import apps
from chemtrails.conf import settings
import csv
import tempfile
from neomodel import db, exception


class Command(BaseCommand):
    help = 'imports current database to neo4j'

    def add_arguments(self, parser):
        pass

    def handle(self, *args, **options):

        target_file = tempfile.NamedTemporaryFile(delete=True)
        BUFFER_COUNT = 100

        try:
            for app in apps.all_models:
                self.stdout.write(self.style.SUCCESS('Looking at {}'.format(app)))
                model_count = 0
                cntr = 0

                for n, model in enumerate(apps.get_app_config(app_label=app).get_models()):
                    model_count += 1
                    cls = get_node_class_for_model(model)
                    if cls._is_ignored:
                        continue

                    for item in model.objects.all():
                        node = cls(instance=item, bind=False)
                        node.to_csv(cntr=cntr, target_file=target_file)
                        cntr += 1

            with open(target_file.name, newline='') as csvfile:
                spamreader = csv.reader(csvfile, delimiter=';', quotechar='|')
                self.stdout.write(self.style.SUCCESS('Crucifixion party! Wait for it...'))
                node_count = 0
                skip_count = 0
                cypher = ''

                for row in spamreader:
                    if row[0] == 'n':
                        try:
                            cypher += row[1]
                            node_count += 1
                        except exception.UniqueProperty:
                            skip_count += 1

                    if cypher and node_count % BUFFER_COUNT == 0:
                        db.cypher_query(cypher)
                        cypher = ''
                        self.stdout.write('{} nodes processed'.format(node_count), ending='\r')
                        self.stdout.flush()

                if cypher:
                    db.cypher_query(cypher)

                self.stdout.write(self.style.SUCCESS(
                    '{} nodes processed and delivered to Neo... '
                    'Why oh why, didn\'t I take the blue pill??'.format(node_count)))
                if skip_count > 0:
                    self.stdout.write(self.style.SUCCESS(
                        '{} nodes already existed, and was not updated.... Were you listening to me Neo? Or were you '
                        'looking at the woman in the red dress?'.format(skip_count)))

            with open(target_file.name, newline='') as csvfile:
                spamreader = csv.reader(csvfile, delimiter=';', quotechar='|')
                relation_count = 0

                for row in spamreader:
                    if row[0] == 'r':
                        db.cypher_query(row[1])
                        relation_count += 1

                    if relation_count % BUFFER_COUNT == 0:
                        self.stdout.write('{} relations processed'.format(relation_count), ending='\r')
                        self.stdout.flush()

                self.stdout.write(self.style.SUCCESS(
                    '{} relations processed and delivered to Neo... '
                    'Why oh why, didn\'t I take the blue pill??'.format(relation_count)))
                self.stdout.write(self.style.SUCCESS('Crucifixion party! By the left! Forward!'))

        finally:
            target_file.close()




# -*- coding: utf-8 -*-

import sys
from neomodel import install_labels

from chemtrails.conf import settings
from chemtrails.neoutils import (
    get_meta_node_for_model, get_meta_node_class_for_model,
    get_node_for_object, get_node_class_for_model, get_nodeset_for_queryset
)


def post_migrate_handler(sender, **kwargs):
    """
    Create the meta graph after migrations has been installed.
    """
    if settings.ENABLED:
        for model in sender.models.values():
            install_labels(get_node_class_for_model(model), quiet=False, stdout=sys.stdout)
            install_labels(get_meta_node_class_for_model(model), quiet=False, stdout=sys.stdout)

            get_meta_node_for_model(model).sync(max_depth=settings.MAX_CONNECTION_DEPTH, update_existing=True)


def post_save_handler(sender, instance, **kwargs):
    """
    Sync the node instance after it has been saved.
    """
    if settings.ENABLED:
        get_node_for_object(instance, bind=False).sync(max_depth=settings.MAX_CONNECTION_DEPTH, update_existing=True)


def pre_delete_handler(sender, instance, **kwargs):
    """
    Delete the node from the graph before it is removed from the database.
    """
    if settings.ENABLED:
        klass = get_node_class_for_model(instance._meta.model)
        node = klass.nodes.get_or_none(**{'pk': instance.pk})
        if node:
            node.delete()


def m2m_changed_handler(sender, instance, action, reverse, model, pk_set, **kwargs):
    """
    Update relations for the node when m2m relationships are added or deleted.
    """
    if settings.ENABLED:
        if action not in ('post_add', 'post_remove', 'post_clear'):
            return

        if action == 'post_add':
            get_nodeset_for_queryset(model.objects.filter(pk__in=pk_set), sync=True,
                                     max_depth=settings.MAX_CONNECTION_DEPTH)
        elif action == 'post_remove':
            get_nodeset_for_queryset(model.objects.filter(pk__in=pk_set), sync=True,
                                     max_depth=settings.MAX_CONNECTION_DEPTH)
        elif action == 'post_clear':
            get_node_for_object(instance, bind=False).sync(max_depth=settings.MAX_CONNECTION_DEPTH, update_existing=True)



# -*- coding: utf-8 -*-

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class ChemtrailsPermissionsConfig(AppConfig):
    name = 'chemtrails.contrib.permissions'
    verbose_name = _('Graph based permissions')
    label = 'chemtrails_permissions'



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-05-10 13:14
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0004_book_view_book_permission'),
    ]

    operations = [
        migrations.CreateModel(
            name='Guild',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('contact', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='guild_contacts', to='testapp.Author')),
                ('members', models.ManyToManyField(related_name='guild_set', to='testapp.Author', verbose_name='members')),
            ],
        ),
        migrations.AddField(
            model_name='author',
            name='guilds',
            field=models.ManyToManyField(blank=True, to='testapp.Guild'),
        ),
    ]



# -*- coding: utf-8 -*-

import json

from collections import OrderedDict
from operator import itemgetter

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import ugettext_lazy as _

from neomodel.match import OUTGOING, INCOMING, EITHER

from chemtrails.contrib.permissions.forms import JSONField
from chemtrails.neoutils.query import get_node_relationship_types, get_node_permissions, get_relationship_types


def get_node_relations_choices():
    mapping = [(v, k) for k, v in get_node_relationship_types(params={'type': 'MetaNode'}).items()]
    return sorted(mapping, key=itemgetter(0))


def get_node_permissions_choices():
    mapping = [(v, _('%s' % v).capitalize()) for v in get_node_permissions()]
    return sorted(mapping, key=itemgetter(0))


def get_relationship_types_choices():
    mapping = [(v, _('%s' % v).upper()) for v in get_relationship_types()]
    return sorted(mapping, key=itemgetter(0))


class AccessRule(models.Model):

    DIRECTION_CHOICES = (
        (INCOMING, _('INCOMING')),
        (EITHER, _('EITHER')),
        (OUTGOING, _('OUTGOING'))
    )

    ctype_source = models.ForeignKey(ContentType, verbose_name=_('source content type'),
                                     related_name='accessrule_ctype_source_set', on_delete=models.CASCADE)
    ctype_target = models.ForeignKey(ContentType, verbose_name=_('target content type'),
                                     related_name='accessrule_ctype_target_set', on_delete=models.CASCADE)
    description = models.TextField(_('description'), default='')
    permissions = models.ManyToManyField(Permission, verbose_name=_('access rule permissions'), blank=True,
                                         help_text=_('Required permissions for target node.'),
                                         related_name='accessrule_permissions', related_query_name='accessrule')
    relation_types = ArrayField(verbose_name=_('relation types'), base_field=JSONField(), default=list,
                                help_text=_('List of relation type rule definitions, optionally with a map '
                                            'of properties for matching the relation type node. '
                                            'Example: {"USER": {"is_superuser": true}}, {"GROUP": null}'))
    direction = models.SmallIntegerField(verbose_name=_('direction'), choices=DIRECTION_CHOICES,
                                         blank=True, null=True, default=None,
                                         help_text=_('Override calculated direction for relationships.<br/> '
                                                     'Use with caution, this will affect *all* directional '
                                                     'relationships for evaluation of this rule!'))
    is_active = models.BooleanField(default=True, help_text=_('Uncheck to disable evaluation of the rule '
                                                              'in the rule chain.'))
    requires_staff = models.BooleanField(default=False, help_text=_('Requires user which should have '
                                                                    'permissions evaluated to be a staff user.'))
    created = models.DateTimeField(verbose_name=_('created'), auto_now_add=True)
    updated = models.DateTimeField(verbose_name=_('updated'), auto_now=True)

    class Meta:
        ordering = ('ctype_target', '-created',)

    def __repr__(self):
        return '<%(class)s: [%(source)s]-[*]-[%(target)s]>' % {
            'class': self.__class__.__name__,
            'source': '%s.%s' % (self.ctype_source.app_label, self.ctype_source.model),
            'target': '%s.%s' % (self.ctype_target.app_label, self.ctype_target.model),
        }

    def __str__(self):
        return '[%(source)s]-[*]-[%(target)s]' % {
            'source': self.ctype_source,
            'target': self.ctype_target
        }

    @property
    def relation_types_obj(self):
        """
        Return the relation types JSON string as an OrderedDict.
        """
        if self.relation_types:
            return [json.loads(rule, object_pairs_hook=OrderedDict) for rule in self.relation_types]
        else:
            return OrderedDict()



# -*- coding: utf-8 -*-

from collections import defaultdict

from django.core.exceptions import ValidationError

from libcypher_parser import parse_statement
from neomodel import db
from chemtrails.utils import flatten


def get_relationship_types():
    """
    :returns: A flat list of all relationship types in the database.
    :rtype: list
    """
    query = 'MATCH (n)-[r]-() RETURN DISTINCT type(r) ORDER BY (type(r))'
    validate_cypher(query, raise_exception=True)
    result, _ = db.cypher_query(query)
    return list(flatten(result))


def get_node_permissions():
    """
    :returns: A distinct list of all available permissions.
    :rtype: list
    """
    query = 'MATCH (n)-[r]-() WHERE n.type = \'MetaNode\' RETURN DISTINCT n.default_permissions'
    validate_cypher(query, raise_exception=True)
    result, _ = db.cypher_query(query)
    return list(flatten(result))


def get_node_relationship_types(params=None):
    """
    :param: params: Optional node filtering parameters.
    :returns: Dictionary with node label and relationship types.
    :rtype: dict[list]
    """
    mapping = defaultdict(list)
    query = ' '.join(('MATCH (n)-[r]-()',
                      'WHERE' if params else '',
                      ' AND '.join(['n.{} = {{{}}}'.format(key, key) for key in params.keys()]) if params else '',
                      'RETURN DISTINCT labels(n), type(r)'))
    validate_cypher(query, raise_exception=True)
    result, _ = db.cypher_query(query, params)
    for label, relationship_type in result:
        label = label[0]
        mapping[label].append(relationship_type)
    return mapping


def validate_cypher(statement, raise_exception=False, exc_class=ValidationError):
    """
    Checks if ``statement`` is a valid Cypher statement.
    
    :param statement: Cypher statement to parse.
    :type statement: str
    :param raise_exception: Raise ValidationError if not valid.
    :type raise_exception: bool
    :param exc_class: Exception class to raise
    :type exc_class: Class
    """
    is_valid, data = parse_statement(statement)
    if not is_valid and raise_exception:
        raise exc_class('Failed to validate Cypher statement', params=data)
    return is_valid, data



# -*- coding: utf-8 -*-

from chemtrails.compat import rest_framework
assert rest_framework, ('chemtrails.contrib.permissions module requires '
                        'django-rest-framework to be installed. Install with '
                        '`pip install djangorestframework`.')

default_app_config = 'chemtrails.contrib.permissions.apps.ChemtrailsPermissionsConfig'



# -*- coding: utf-8 -*-

import re
import inspect

from django.contrib.contenttypes.fields import GenericForeignKey
from neomodel.match import Traversal
from neomodel.match import OUTGOING, INCOMING, EITHER


def build_relation_string(lhs, rhs, ident=None, relation_type=None, direction=None, props=None, **kwargs):
    """
    Generate a relationship matching string, with specified parameters.
    
    Basically a rewrite of ``neomodel.match._rel_helper`` except that it 
    returns a string suitable for doing string formatting on.
    
    :param lhs: The left hand statement.
    :type lhs: str
    :param rhs: The right hand statement.
    :type rhs: str
    :param ident: A specific identity to name the relationship, or None.
    :type ident: str
    :param relation_type: None for all direct rels, * for all of any length, or a name of an explicit rel.
    :type relation_type: str
    :param direction: None or EITHER for all OUTGOING,INCOMING,EITHER. Otherwise OUTGOING or INCOMING.
    :param props: dictionary of relationship properties to match
    :returns: The calculated relationship string
    :rtype str
    """
    direction_map = {
        OUTGOING: '-{0}->',
        INCOMING: '<-{0}-',
        EITHER: '-{0}-'
    }

    props = props or {}
    relation_props = ' {{{{{0}}}}}'.format(', '.join(
        ['{}: {}'.format(key, value) for key, value in props.items()]))

    statement = direction_map[direction or 0]
    if relation_type is None:
        statement = statement.format('')
    elif relation_type == '*':
        statement = statement.format('[*]')
    else:
        statement = statement.format('[%s:%s%s]' % (ident if ident else '', relation_type, relation_props))
    return '({0}){1}({2})'.format(lhs, statement, rhs)


class PathManager:
    """
    Class for building traversal paths for node instances
    managed by chemtrails.
    """
    def __init__(self, source):
        self.source = source

        # Keep a list of all calculated matching strings.
        self._statements = []

        # Keep track of the next source class.
        self.___next_class__ = source

        # Override default direction. If set to -1, 0 or 1 this
        # will affect **all** directions in the query!
        self.__direction__ = None

    @property
    def traversals(self):
        """
        Returns a list of valid ``Traversal`` instances
        for the current nodeset.
        """
        return list(filter(lambda prop: isinstance(prop, Traversal),
                           self.next_class.nodes.__dict__.values()))

    @property
    def next_class(self):
        return self.___next_class__

    @next_class.setter
    def next_class(self, klass):
        self.___next_class__ = klass

    @property
    def direction(self):
        return self.__direction__

    @direction.setter
    def direction(self, direction):
        self.__direction__ = direction

    @property
    def statement(self):
        """
        :returns: The final calculated relationship statement or None if no
                  relation types has been added.
        :rtype: None or str
        """
        if not self._statements:
            return None

        def format_node(ident, label, **filters):
            separator = ': ' if label else ''
            if not label and not filters:
                return '{0}'.format(ident)
            elif not filters:
                return '{0}: {1}'.format(ident, label)
            else:
                label = '{0} {{{1}}}'.format(
                    label if label else '', ', '.join(['{}: {}'.format(
                        key, '"%s"' % value if isinstance(value, str) else value)
                        for key, value in filters.items()])
                )
                return '{0}{1}{2}'.format(ident, separator, label)

        # Matches ie. (source1: UserNode {...}) as long as it's followed
        # by a "-[" or "<-[" which indicates the beginning of a relationship.
        pattern = re.compile(r'^(\(source\d+:.\w+(.{\w+:.[A-Za-z0-9\'"_]+})?\)(?=<?-\[))')

        statements = []
        for n, config in enumerate(self._statements):
            # Replace placeholders with actual values.
            defaults = config['relation_props'].copy()

            # Check if we should override direction
            if self.direction is not None:
                assert self.direction in (INCOMING, EITHER, OUTGOING), (
                    'Direction must be one of %s.' % ', '.join((INCOMING, EITHER, OUTGOING)))
                config['traversal'].definition['direction'] = self.direction

            atom = build_relation_string(lhs='{source}', rhs='{target}',
                                         props={key: '{{{0}}}'.format(key) for key in defaults.keys()},
                                         **config['traversal'].definition)

            source_props = self.resolve_filters(config['source_props'])
            if not inspect.isclass(config['source_class']):
                # If we have a node instance, always match its primary key!
                if getattr(config['source_class'], 'pk', None):
                    source_props['pk'] = config['source_class'].pk

            target_props = self.resolve_filters(config['target_props'])

            defaults.update({
                'source': 'source{0}'.format(format_node(
                    ident=n,
                    label=(config['source_class'].__label__
                           if config['source_class'] else None),
                    **source_props
                )),
                'target': 'target{0}'.format(format_node(
                    ident=config.get('target_index', n),
                    label=(config['target_class'].__label__
                           if config['target_class'] else None),
                    # Add any user specified filters to target node.
                    **target_props
                ))
            })

            relation_str = atom.format(**defaults)
            if n == 0:
                statements.append(relation_str)
                continue

            # Remove the source node definition from string.
            # It's already specified in the previous element.
            statements.append(pattern.sub('', relation_str))

        return ''.join(statements)

    def add(self, relation_type=None, relation_props=None, source_props=None, target_props=None):
        """
        Adds a relationship matching string, based on relation type.
        
        :param relation_type: The relationship type, ie USER.
        :type relation_type: str
        :param relation_props: Optional properties used to instantiate the
          ``StructuredRel`` relationship class. This is used to gather properties 
          for the generated relationship statement.
        :type relation_props: dict
        :param source_props: Property mapping which should be applied for filtering the
          source node.
        :type source_props: dict
        :param target_props: Property mapping which should be applied for filtering the
          target node.
        :type target_props: dict
        :returns: self
        """
        defaults = {}
        traversal = self.get_traversal(relation_type)
        if traversal is None:
            # Check if using any back-references like ie. {0:GROUPS}
            pattern = re.compile(r'(?<=^{)(\d:\w+)(?=})')
            match = pattern.search(relation_type)
            if match:
                index, relation_type = match.group().split(':')
                index = int(index)
                traversal = self.get_traversal(relation_type)
                if len(self._statements) - 1 < index:
                    raise IndexError('target index out of range')
                defaults['target_index'] = index
            elif not relation_type:
                raise ValueError('Cannot find relationship with empty relation type.')
            if traversal is None:
                raise AttributeError('%(klass)r has no relation type %(relation_type)s' % {
                    'klass': self.next_class,
                    'relation_type': relation_type
                })

        model = traversal.definition['model']
        relation_props = relation_props or {}

        # Instantiate a fake relationship model in order
        # to pick attributes for the relationship.
        fake = model(**relation_props)
        relation_props = {prop: '"%s"' % value if isinstance(value, str) else value
                          for prop, value in model.deflate(fake.__properties__).items()}

        # At this point we have no idea of what object the GenericForeignKey
        # relationship is really pointing at.
        remote_field = getattr(traversal.source_class.Meta.model,
                               model.remote_field.default.split('.')[-1])
        if isinstance(remote_field, GenericForeignKey):
            del relation_props['target_field']

        defaults.update({
            'source_class': self.next_class,
            'source_props': source_props or {},
            'target_class': (traversal.target_class
                             if not isinstance(remote_field, GenericForeignKey) else None),
            'target_props': target_props or {},
            'relation_props': relation_props,
            'traversal': traversal,
        })
        self._statements.append(defaults)
        self.next_class = traversal.target_class
        return self

    def get_traversal(self, relation_type):
        """
        Returns the first traversal instance matching ``relation_type``.
        """
        for traversal in self.traversals:
            if traversal.definition['relation_type'] == relation_type:
                return traversal
        return None
    
    def get_path(self):
        """
        :returns: The calculated statement as MATCH path = (...) RETURN path
        :rtype str
        """
        if not self.statement:
            raise AttributeError('No calculated statements.')
        return 'MATCH path = {statement} RETURN path;'.format(statement=self.statement)

    def get_match(self):
        if not self.statement:
            raise AttributeError('No calculated statements.')
        return 'MATCH {statement} RETURN *;'.format(statement=self.statement)

    def resolve_filters(self, filters):
        """
        Check if the filter contains any special variables and resolve them.
        Currently only supports resolving parameters using "{source}" variable,
        which will inspect the originating node.

        Ex:
          filters = {'username': '{source}.username'}

          This will replace the '{source}.pk' value with whatever value is found
          on the 'username' attribute on the source node.

        :param filters: Mapping with filter to inspect
        :type filters: dict
        :returns: Dictionary with resolved values.
        """
        pattern = re.compile(r'(?<={source}.)\w+')  # Matches '{source}.attribute'
        for attr, value in filters.items():
            try:
                match = pattern.search(value)
                if match:
                    key = match.group()
                    if key not in self.source.defined_properties(aliases=False, rels=False):
                        raise AttributeError("%(node)r has no valid property named '%(key)s'. "
                                             "Make sure the '%(value)s' targets a valid attribute." % {
                                                 'node': self.source, 'key': key, 'value': value
                                             })
                    value = getattr(self.source, key)
                    if value is None:
                        value = 'null'
                    filters[attr] = value
            except TypeError:
                # This can happen if trying to re-process an already processed filter.
                continue

        return filters



# -*- coding: utf-8 -*-

from django.contrib.auth import get_user_model
from django.core import serializers
from django.test import TestCase

from chemtrails.contrib.permissions.models import AccessRule
from chemtrails.contrib.permissions.utils import get_content_type
from tests.utils import TestCaseMixins

USER_MODEL = get_user_model()


class JSONFieldTestCase(TestCase, TestCaseMixins):

    def test_field_is_json_serializable(self):
        instance = AccessRule.objects.create(ctype_source=get_content_type(USER_MODEL),
                                             ctype_target=get_content_type(USER_MODEL),
                                             relation_types=[
                                                 {'GROUPS': None}
                                             ])
        json = serializers.serialize('json', [instance])
        self.assertIsJSON(json)




# -*- coding: utf-8 -*-

import ast
import functools
import logging
import os
import time
from collections import Sequence

logger = logging.getLogger(__name__)


def get_model_string(model):
    """
    :param model: model
    :returns: <app_label>.<model_name> string representation for the model
    """
    return "{app_label}.{model_name}".format(app_label=model._meta.app_label, model_name=model._meta.model_name)


def flatten(sequence):
    """
    Flatten an arbitrary nested sequence.
    Example usage:
      >> my_list = list(flatten(nested_lists))
    :param sequence: A nested list or tuple.
    :returns: A generator with all values in a flat structure.
    """
    for i in sequence:
        if isinstance(i, Sequence) and not isinstance(i, (str, bytes)):
            yield from flatten(i)
        else:
            yield i


def timeit(func):
    """
    Decorator which logs the timing of executing a function
    or method call to DEBUG logger.
    """
    @functools.wraps(func)
    def f(*args, **kwargs):
        timestamp = time.time()
        func(*args, **kwargs)
        logger.debug('function [{0}] executed in {1}s'.format(
            f.__name__, '%.5f' % (time.time() - timestamp)
        ))
    return f


def get_environment_variable(name, fallback=None):
    """
    Get environment variable and try to cast it to proper python type.
    :param name: Environment variable name to look for.
    :param fallback: Default fallback value if ``name`` is not found.
    :returns: Environment variable value if found, else None.
    """
    value = os.environ.get(name, fallback)
    try:
        if value in ('true', 'TRUE', 'false', 'FALSE'):
            value = value.capitalize()
        value = ast.literal_eval(value)
    except ValueError:
        pass
    return value




# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-06-09 09:26
from __future__ import unicode_literals

import chemtrails.contrib.permissions.forms.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0005_guild'),
    ]

    operations = [
        migrations.CreateModel(
            name='JSONModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('json', chemtrails.contrib.permissions.forms.fields.JSONField(default=dict)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-

import re
import os

from pip.req import parse_requirements
from setuptools import setup, find_packages


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


setup(
    name='django-chemtrails',
    version=get_version('chemtrails'),
    description='Graphing Django in Neo4j',
    long_description='Graphing Django in Neo4j slightly longer.',
    author='Rolf Håvard Blindheim',
    author_email='rhblind@gmail.com',
    url='https://github.com/inonit/django-chemtrails',
    download_url='https://github.com/inonit/django-chemtrails.git',
    license='MIT License',
    packages=find_packages('.', exclude=['tests*', 'docs*', 'javascript*']),
    include_package_data=True,
    install_requires=[str(r.req) for r in parse_requirements('./requirements.txt', session=False)],
    extras_require={
        'perms': [
            'djangorestframework'
        ]
    },
    tests_require=[
        'nose',
        'coverage',
        'django-autofixture',
        'djangorestframework',
        'psycopg2>=2.6.2'
    ],
    zip_safe=False,
    test_suite='tests.run_tests.start',
    classifiers=[
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-06-15 06:45
from __future__ import unicode_literals

import chemtrails.contrib.permissions.forms.fields
import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('auth', '0008_alter_user_username_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='AccessRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(default='', verbose_name='description')),
                ('relation_types', django.contrib.postgres.fields.ArrayField(base_field=chemtrails.contrib.permissions.forms.fields.JSONField(default={}), default=list, help_text='List of relation type rule definitions, optionally with a map of properties for matching the relation type node. Example: {"USER": {"is_superuser": true}}, {"GROUP": null}', size=None, verbose_name='relation types')),
                ('is_active', models.BooleanField(default=True, help_text='Uncheck to disable evaluation of the rule in the rule chain.')),
                ('requires_staff', models.BooleanField(default=False, help_text='Requires user which should have permissions evaluated to be a staff user.')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
                ('ctype_source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accessrule_ctype_source_set', to='contenttypes.ContentType', verbose_name='source content type')),
                ('ctype_target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accessrule_ctype_target_set', to='contenttypes.ContentType', verbose_name='target content type')),
                ('permissions', models.ManyToManyField(blank=True, help_text='Required permissions for target node.', related_name='accessrule_permissions', related_query_name='accessrule', to='auth.Permission', verbose_name='access rule permissions')),
            ],
            options={
                'ordering': ('ctype_target', '-created'),
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-10-04 23:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('chemtrails_permissions', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='accessrule',
            name='direction',
            field=models.SmallIntegerField(blank=True, choices=[(-1, 'INCOMING'), (0, 'EITHER'), (1, 'OUTGOING')], default=None, help_text='Override calculated direction for relationships.<br/> Use with caution, this will affect *all* directional relationships for evaluation of this rule!', null=True, verbose_name='direction'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-06-12 09:27
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0006_jsonmodel'),
    ]

    operations = [
        migrations.DeleteModel(
            name='JSONModel',
        ),
    ]



# -*- coding: utf-8 -*-

from collections import OrderedDict

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.utils import six
from django.utils.translation import ugettext_lazy as _

from neomodel import properties
from neomodel.cardinality import One
from neomodel.relationship import RelationshipMeta

from rest_framework import serializers

from chemtrails.contrib.permissions.models import AccessRule

_field_mapping = {
    properties.AliasProperty: lambda prop: _field_mapping[prop.target],
    properties.ArrayProperty: serializers.ListField,
    properties.BooleanProperty: serializers.BooleanField,
    properties.DateProperty: serializers.DateField,
    properties.DateTimeProperty: serializers.DateTimeField,
    properties.EmailProperty: serializers.EmailField,
    properties.FloatProperty: serializers.FloatField,
    properties.IntegerProperty: serializers.IntegerField,
    properties.JSONProperty: serializers.JSONField,
    properties.RegexProperty: serializers.RegexField,
    properties.StringProperty: serializers.CharField,
    properties.UniqueIdProperty: serializers.UUIDField
}


class RelationshipSerializer(serializers.Serializer):
    """
    REST Framework based Serializer which dynamically
    serializes a ``StructuredRel`` instance.
    """
    def get_fields(self):
        field_mapping = OrderedDict()
        for key, value in self.instance.items():
            if key == 'node_class':
                field_mapping['to'] = serializers.CharField(default=value.__label__)
            elif isinstance(value, RelationshipMeta):
                field_mapping['meta'] = serializers.DictField(
                    child=serializers.CharField(),
                    default={k: v.default for k, v in value.__dict__.items() if isinstance(v, properties.Property)})
            elif key == 'direction':
                field_mapping[key] = serializers.IntegerField(default=value)
            elif key == 'relation_type':
                field_mapping[key] = serializers.CharField(default=value)
        return field_mapping


class NodeSerializer(serializers.Serializer):
    """
    REST Framework based Serializer which dynamically
    serializes a ``StructuredNode`` instance.
    """
    def get_fields(self):
        field_mapping = OrderedDict()

        # Normal properties
        for field, property_class in self.instance.defined_properties(aliases=False, rels=False).items():
            # TODO: Support AliasField - check for __call__ in _field_mapping.
            field_mapping['id'] = serializers.IntegerField(default=self.instance.id)
            field_mapping['label'] = serializers.CharField(default=self.instance.__label__)
            field_mapping.update({
                field: self.get_serializer_field(property_class, **self._get_default_field_kwargs(property_class))
            })

        # Relationships
        for field, property_class in self.instance.defined_properties(aliases=False, properties=False).items():
            field_mapping.update({
                field: RelationshipSerializer(property_class.definition, many=not isinstance(property_class, One))
            })

        return field_mapping
    
    @staticmethod
    def _get_default_field_kwargs(property_class):
        """
        Return default kwargs used to initialize
        the Serializer field with.
        :param property_class: Neomodel property class
        """
        defaults = property_class.__dict__.copy()
        delete_attrs = [
            'db_property',
            'choices',
            'has_default',
            'index',
            'unique_index'
        ]
        for attr in delete_attrs:
            if attr in defaults:
                del defaults[attr]

        return defaults
    
    def get_serializer_field(self, property_class, **kwargs):
        """
        Returns the Serializer field instance that should be
        used for validating and deserializing the field.
        :param property_class: Neomodel property class.
        :param kwargs: Mapping with options used to initialize the Serializer field instance.
        """
        field = _field_mapping[property_class.__class__]
        return field(**kwargs)


class ContentTypeIdentityField(serializers.RelatedField):
    """
    REST Framework related field for ContentType objects.
    This field represents related ContentType objects by their natural key.
    """
    default_error_messages = {
        'required': _('This field is required.'),
        'does_not_exist': _('Invalid content type "{ctype}" - object does not exist.'),
        'incorrect_type': _('Incorrect type. Expected content type string identifier, received {data_type}.'),
        'incorrect_length': _('Incorrect length. Expected content type string, '
                              'separated by punctuation. Received "{data}".')
    }

    def to_internal_value(self, data):
        if not isinstance(data, six.text_type):
            self.fail('incorrect_type', data_type=type(data).__name__)

        try:
            app_label, model = data.split('.', 1)
            return ContentType.objects.get_by_natural_key(app_label=app_label, model=model)
        except ContentType.DoesNotExist:
            self.fail('does_not_exist', ctype=data)
        except ValueError:
            self.fail('incorrect_length', data=data)

    def to_representation(self, value):
        return '{app_label}.{model}'.format(app_label=value.app_label, model=value.model)


class PermissionIdentityField(serializers.RelatedField):
    """
    REST Framework related field for Permission objects.
    This field represents related Permission objects by their natural key.
    """
    default_error_messages = {
        'required': _('This field is required.'),
        'does_not_exist': _('Invalid permission "{data}" - object does not exist.'),
        'incorrect_type': _('Incorrect type. Expected permission string identifier, received {data_type}.'),
        'incorrect_length': _('Incorrect length. Expected permission string identifier, '
                              'separated by punctuation. Received "{data}".')
    }

    def to_internal_value(self, data):
        if not isinstance(data, six.text_type):
            self.fail('incorrect_type', data_type=type(data).__name__)

        try:
            app_label, codename = data.split('.', 1)
            return Permission.objects.select_related('content_type').get(
                codename=codename, content_type=ContentType.objects.get(app_label=app_label,
                                                                        permission__codename=codename))
        except (Permission.DoesNotExist, ContentType.DoesNotExist):
            self.fail('does_not_exist', data=data)
        except ValueError:
            self.fail('incorrect_length', data=data)

    def to_representation(self, value):
        codename, app_label, _ = value.natural_key()
        return '{app_label}.{codename}'.format(app_label=app_label, codename=codename)


class AccessRuleSerializer(serializers.ModelSerializer):
    """
    REST Framework Serializer for AccessRule objects.
    """
    ctype_source = ContentTypeIdentityField(queryset=ContentType.objects.all(), required=True)
    ctype_target = ContentTypeIdentityField(queryset=ContentType.objects.all(), required=True)
    permissions = PermissionIdentityField(queryset=Permission.objects.all(), many=True)
    relation_types = serializers.ListField(child=serializers.JSONField())

    class Meta:
        model = AccessRule
        fields = '__all__'
        read_only_fields = ('created', 'updated')
